#!/usr/bin/env -S uv run

# /// script
# dependencies = ["nox>=2025.2.9"]
# ///

"""
cibuildwheel's nox support

Tags:

    lint: All linting jobs
    update: All update jobs

See sessions with `nox -l`
"""

import os
import shutil
import sys
from pathlib import Path

import nox

nox.needs_version = ">=2025.2.9"
nox.options.default_venv_backend = "uv|virtualenv"

DIR = Path(__file__).parent.resolve()


@nox.session(tags=["lint"])
def lint(session: nox.Session) -> None:
    """
    Run the linter.
    """
    session.install("prek")
    session.run("prek", "run", "--all-files", *session.posargs)


@nox.session(tags=["lint"])
def pylint(session: nox.Session) -> None:
    """
    Run pylint.
    """

    session.install("pylint>=3.2", "-e.")
    session.run("pylint", "cibuildwheel", *session.posargs)


@nox.session
def tests(session: nox.Session) -> None:
    """
    Run the unit and regular tests.
    """
    pyproject = nox.project.load_toml()
    session.install("-e.", *nox.project.dependency_groups(pyproject, "test"))
    if session.posargs:
        session.run("pytest", *session.posargs)
    else:
        unit_test_args = ["--run-docker"] if sys.platform.startswith("linux") else []
        session.run("pytest", "cibuildwheel")  # run doctests
        session.run("pytest", "unit_test", *unit_test_args)
        session.run("pytest", "test", "-x", "--durations", "0", "--timeout=2400", "test")


@nox.session(default=False, tags=["update"])
def update_constraints(session: nox.Session) -> None:
    """
    Update the dependencies inplace.
    """

    session.install("-e.", "click")

    resources = Path("cibuildwheel/resources")

    if session.venv_backend != "uv":
        session.install("uv>=0.1.23")

    # CUSTOM_COMPILE_COMMAND is a pip-compile option that tells users how to
    # regenerate the constraints files
    env = os.environ.copy()
    env["UV_CUSTOM_COMPILE_COMMAND"] = f"nox -s {session.name}"

    for minor_version in range(8, 15):
        python_version = f"3.{minor_version}"
        output_file = resources / f"constraints-python{python_version.replace('.', '')}.txt"
        session.run(
            "uv",
            "pip",
            "compile",
            f"--python-version={python_version}",
            "--upgrade",
            resources / "constraints.in",
            f"--output-file={output_file}",
            env=env,
        )

    shutil.copyfile(
        resources / "constraints-python314.txt",
        resources / "constraints.txt",
    )

    build_platforms = nox.project.load_toml(resources / "build-platforms.toml")
    pyodides = build_platforms["pyodide"]["python_configurations"]
    for pyodide in pyodides:
        python_version = ".".join(pyodide["version"].split(".")[:2])
        pyodide_version = pyodide["default_pyodide_version"]

        tmp_file = Path(session.create_tmp()) / "constraints-pyodide.in"

        session.run(
            "python",
            "bin/generate_pyodide_constraints.py",
            "--output-file",
            tmp_file,
            pyodide_version,
        )

        output_file = resources / f"constraints-pyodide{python_version.replace('.', '')}.txt"
        session.run(
            "uv",
            "pip",
            "compile",
            f"--python-version={python_version}",
            "--upgrade",
            tmp_file,
            f"--output-file={output_file}",
            env=env,
        )


@nox.session(default=False, tags=["update"])
def update_pins(session: nox.Session) -> None:
    """
    Update the python, docker, virtualenv, node, and python-build-standalone
    version pins inplace.
    """
    pyproject = nox.project.load_toml()
    session.install("-e.", *nox.project.dependency_groups(pyproject, "bin"))
    session.run("python", "bin/update_pythons.py", "--force")
    session.run("python", "bin/update_docker.py")
    session.run("python", "bin/update_virtualenv.py", "--force")
    session.run("python", "bin/update_nodejs.py", "--force")
    session.run("python", "bin/update_python_build_standalone.py")


@nox.session(default=False, reuse_venv=True, tags=["update"])
def update_proj(session: nox.Session) -> None:
    """
    Update the README inplace.
    """
    session.install_and_run_script(
        "bin/projects.py",
        "docs/data/projects.yml",
        *session.posargs,
    )


@nox.session(default=False, reuse_venv=True, tags=["update"])
def generate_schema(session: nox.Session) -> None:
    """
    Generate the cibuildwheel.schema.json file.
    """
    output = session.install_and_run_script("bin/generate_schema.py", silent=True)
    assert isinstance(output, str)
    DIR.joinpath("cibuildwheel/resources/cibuildwheel.schema.json").write_text(output)


@nox.session(default=False, reuse_venv=True)
def bump_version(session: nox.Session) -> None:
    """
    Bump cibuildwheel's version. Interactive.
    """
    session.install_and_run_script("bin/bump_version.py")


@nox.session(default=False)
def docs(session: nox.Session) -> None:
    """
    Build the docs. Will serve unless --non-interactive
    """
    pyproject = nox.project.load_toml()
    session.install("-e.", *nox.project.dependency_groups(pyproject, "docs"))
    session.run("mkdocs", "serve" if session.interactive else "build", "--strict", *session.posargs)


@nox.session(default=False)
def build(session: nox.Session) -> None:
    """
    Build an SDist and wheel.
    """

    build_p = DIR.joinpath("build")
    if build_p.exists():
        shutil.rmtree(build_p)

    dist_p = DIR.joinpath("dist")
    if dist_p.exists():
        shutil.rmtree(dist_p)

    session.install("build")
    session.run("python", "-m", "build")


if __name__ == "__main__":
    nox.main()



import jinja2

from .base import TestProject
from .c import SPAM_C_TEMPLATE

SETUP_PY_TEMPLATE = r"""
import os
import sys

from setuptools import setup, Extension

{{ setup_py_add }}

libraries = []
# Emscripten fails if you pass -lc...
# See: https://github.com/emscripten-core/emscripten/issues/16680
if sys.platform.startswith('linux') and "emscripten" not in os.environ.get("_PYTHON_HOST_PLATFORM", ""):
    libraries.extend(['m', 'c'])


setup(
    ext_modules=[Extension(
        'spam',
        sources=['spam.c'],
        libraries=libraries,
        {{ setup_py_extension_args_add | indent(8) }}
    )],
    {{ setup_py_setup_args_add | indent(4) }}
)
"""

SETUP_CFG_TEMPLATE = r"""
[metadata]
name = spam
version = 0.1.0

{{ setup_cfg_add }}
"""


def new_c_project(
    *,
    spam_c_top_level_add: str = "",
    spam_c_function_add: str = "",
    setup_py_add: str = "",
    setup_py_extension_args_add: str = "",
    setup_py_setup_args_add: str = "",
    setup_cfg_add: str = "",
) -> TestProject:
    project = TestProject()

    project.files.update(
        {
            "spam.c": jinja2.Template(SPAM_C_TEMPLATE),
            "setup.py": jinja2.Template(SETUP_PY_TEMPLATE),
            "setup.cfg": jinja2.Template(SETUP_CFG_TEMPLATE),
        }
    )

    project.template_context.update(
        {
            "spam_c_top_level_add": spam_c_top_level_add,
            "spam_c_function_add": spam_c_function_add,
            "setup_py_add": setup_py_add,
            "setup_py_extension_args_add": setup_py_extension_args_add,
            "setup_py_setup_args_add": setup_py_setup_args_add,
            "setup_cfg_add": setup_cfg_add,
        }
    )

    return project



import dataclasses
import functools
import json
import os
import shutil
import subprocess
import sys
import tomllib
import typing
from collections.abc import Set
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Final, TypedDict

from filelock import FileLock

from cibuildwheel import errors
from cibuildwheel.architecture import Architecture
from cibuildwheel.environment import ParsedEnvironment
from cibuildwheel.frontend import get_build_frontend_extra_flags
from cibuildwheel.logger import log
from cibuildwheel.options import Options
from cibuildwheel.selector import BuildSelector
from cibuildwheel.util import resources
from cibuildwheel.util.cmd import call, shell
from cibuildwheel.util.file import (
    CIBW_CACHE_PATH,
    copy_test_sources,
    download,
    extract_tar,
    extract_zip,
    move_file,
)
from cibuildwheel.util.helpers import prepare_command, unwrap, unwrap_preserving_paragraphs
from cibuildwheel.util.packaging import find_compatible_wheel, get_pip_version
from cibuildwheel.util.python_build_standalone import (
    PythonBuildStandaloneError,
    create_python_build_standalone_environment,
)
from cibuildwheel.venv import constraint_flags, virtualenv

IS_WIN: Final[bool] = sys.platform.startswith("win")


@dataclasses.dataclass(frozen=True, kw_only=True)
class PythonConfiguration:
    version: str
    identifier: str
    default_pyodide_version: str
    node_version: str


class PyodideXBuildEnvInfoVersionRange(TypedDict):
    min: str | None
    max: str | None


class PyodideXBuildEnvInfo(TypedDict):
    version: str
    python: str
    emscripten: str
    pyodide_build: PyodideXBuildEnvInfoVersionRange
    compatible: bool


@functools.cache
def ensure_node(major_version: str) -> Path:
    with resources.NODEJS.open("rb") as f:
        loaded_file = tomllib.load(f)
    version = str(loaded_file[major_version])
    base_url = str(loaded_file["url"])
    ext = "zip" if IS_WIN else "tar.xz"
    platform = "win" if IS_WIN else ("darwin" if sys.platform.startswith("darwin") else "linux")
    linux_arch = Architecture.native_arch("linux")
    assert linux_arch is not None
    arch = {"x86_64": "x64", "i686": "x86", "aarch64": "arm64"}.get(
        linux_arch.value, linux_arch.value
    )
    name = f"node-{version}-{platform}-{arch}"
    path = CIBW_CACHE_PATH / name
    with FileLock(str(path) + ".lock"):
        if not path.exists():
            url = f"{base_url}{version}/{name}.{ext}"
            with TemporaryDirectory() as tmp_path:
                archive = Path(tmp_path) / f"{name}.{ext}"
                download(url, archive)
                if ext == "zip":
                    extract_zip(archive, path.parent)
                else:
                    extract_tar(archive, path.parent)
    assert path.exists()
    if not IS_WIN:
        return path / "bin"
    return path


def install_emscripten(tmp: Path, version: str) -> Path:
    url = f"https://github.com/emscripten-core/emsdk/archive/refs/tags/{version}.zip"
    installation_path = CIBW_CACHE_PATH / f"emsdk-{version}"
    emsdk_path = installation_path / f"emsdk-{version}/emsdk"
    emcc_path = installation_path / f"emsdk-{version}/upstream/emscripten/emcc"
    with FileLock(f"{installation_path}.lock"):
        if installation_path.exists():
            return emcc_path
        emsdk_zip = tmp / "emsdk.zip"
        download(url, emsdk_zip)
        installation_path.mkdir()
        extract_zip(emsdk_zip, installation_path)
        call(emsdk_path, "install", version)
        call(emsdk_path, "activate", version)

    return emcc_path


def get_all_xbuildenv_version_info(env: dict[str, str]) -> list[PyodideXBuildEnvInfo]:
    xbuildenvs_info_str = call(
        "pyodide",
        "xbuildenv",
        "search",
        "--json",
        "--all",
        env=env,
        cwd=CIBW_CACHE_PATH,
        capture_stdout=True,
    ).strip()

    xbuildenvs_info = json.loads(xbuildenvs_info_str)

    if "environments" not in xbuildenvs_info:
        msg = f"Invalid xbuildenvs info, got {xbuildenvs_info}"
        raise ValueError(msg)

    return typing.cast("list[PyodideXBuildEnvInfo]", xbuildenvs_info["environments"])


def get_xbuildenv_version_info(
    env: dict[str, str], version: str, pyodide_build_version: str
) -> PyodideXBuildEnvInfo:
    xbuildenvs_info = get_all_xbuildenv_version_info(env)
    for xbuildenv_info in xbuildenvs_info:
        if xbuildenv_info["version"] == version:
            return xbuildenv_info

    msg = unwrap(f"""
        Could not find Pyodide cross-build environment version {version} in the available
        versions as reported by pyodide-build v{pyodide_build_version}.
        Available pyodide xbuildenv versions are:
        {", ".join(e["version"] for e in xbuildenvs_info if e["compatible"])}
    """)
    raise errors.FatalError(msg)


# The default pyodide xbuildenv version that's specified in
# build-platforms.toml is compatible with the pyodide-build version that's
# pinned in the bundled constraints file. But if the user changes
# pyodide-version and/or dependency-constraints in the cibuildwheel config, we
# need to check if the xbuildenv version is compatible with the pyodide-build
# version.
def validate_pyodide_build_version(
    xbuildenv_info: PyodideXBuildEnvInfo, pyodide_build_version: str
) -> None:
    """
    Validate the Pyodide version is compatible with the installed
    pyodide-build version.
    """

    pyodide_version = xbuildenv_info["version"]

    if not xbuildenv_info["compatible"]:
        msg = unwrap_preserving_paragraphs(f"""
            The Pyodide xbuildenv version {pyodide_version} is not compatible
            with the pyodide-build version {pyodide_build_version}. Please use
            the 'pyodide xbuildenv search --all' command to find a compatible
            version.

            Set the pyodide-build version using the `dependency-constraints`
            option, or set the Pyodide xbuildenv version using the
            `pyodide-version` option.
        """)
        raise errors.FatalError(msg)


def install_xbuildenv(env: dict[str, str], pyodide_build_version: str, pyodide_version: str) -> str:
    """Install a particular Pyodide xbuildenv version and set a path to the Pyodide root."""
    # Since pyodide-build was unvendored from Pyodide v0.27.0, the versions of
    # pyodide-build are uncoupled from the versions of Pyodide. So, we specify
    # both the pyodide-build version and the Pyodide version in the temp path.
    xbuildenv_cache_path = CIBW_CACHE_PATH / f"pyodide-build-{pyodide_build_version}"
    pyodide_root = xbuildenv_cache_path / pyodide_version / "xbuildenv" / "pyodide-root"

    with FileLock(CIBW_CACHE_PATH / "xbuildenv.lock"):
        if pyodide_root.exists():
            return str(pyodide_root)

        # We don't want to mutate env but we need to delete any existing
        # PYODIDE_ROOT so copy it first.
        env = dict(env)
        env.pop("PYODIDE_ROOT", None)

        # Install the xbuildenv
        call(
            "pyodide",
            "xbuildenv",
            "install",
            "--path",
            str(xbuildenv_cache_path),
            pyodide_version,
            env=env,
            cwd=CIBW_CACHE_PATH,
        )
        assert pyodide_root.exists()

    return str(pyodide_root)


def get_base_python(tmp: Path, python_configuration: PythonConfiguration) -> Path:
    try:
        return create_python_build_standalone_environment(
            python_version=python_configuration.version,
            temp_dir=tmp,
            cache_dir=CIBW_CACHE_PATH,
        )
    except PythonBuildStandaloneError as e:
        msg = unwrap(f"""
            Failed to create a Python build environment:
            {e}
        """)
        raise errors.FatalError(msg) from e


def setup_python(
    tmp: Path,
    python_configuration: PythonConfiguration,
    constraints_path: Path | None,
    environment: ParsedEnvironment,
    user_pyodide_version: str | None,
) -> dict[str, str]:
    log.step("Installing a base python environment...")
    base_python = get_base_python(tmp / "base", python_configuration)

    log.step("Setting up build environment...")
    pyodide_version = user_pyodide_version or python_configuration.default_pyodide_version
    venv_path = tmp / "venv"
    env = virtualenv(python_configuration.version, base_python, venv_path, None, use_uv=False)
    venv_bin_path = venv_path / "bin"
    assert venv_bin_path.exists()
    env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"

    # upgrade pip to the version matching our constraints
    # if necessary, reinstall it to ensure that it's available on PATH as 'pip'
    call(
        "python",
        "-m",
        "pip",
        "install",
        "--upgrade",
        "pip",
        *constraint_flags(constraints_path),
        env=env,
        cwd=venv_path,
    )

    env = environment.as_dictionary(prev_environment=env)

    # check what Python version we're on
    which_python = call("which", "python", env=env, capture_stdout=True).strip()
    print(which_python)
    if which_python != str(venv_bin_path / "python"):
        msg = "python available on PATH doesn't match our venv instance. If you have modified PATH, ensure that you don't overwrite cibuildwheel's entry or insert python above it."
        raise errors.FatalError(msg)
    call("python", "--version", env=env)

    # check what pip version we're on
    assert (venv_bin_path / "pip").exists()
    which_pip = call("which", "pip", env=env, capture_stdout=True).strip()
    print(which_pip)
    if which_pip != str(venv_bin_path / "pip"):
        msg = "pip available on PATH doesn't match our venv instance. If you have modified PATH, ensure that you don't overwrite cibuildwheel's entry or insert pip above it."
        raise errors.FatalError(msg)
    call("pip", "--version", env=env)

    log.step("Installing build tools...")
    call(
        "pip",
        "install",
        "--upgrade",
        "pyodide-build",
        *constraint_flags(constraints_path),
        env=env,
    )

    pyodide_build_version = call(
        "python",
        "-c",
        "from importlib.metadata import version; print(version('pyodide-build'))",
        env=env,
        capture_stdout=True,
    ).strip()

    xbuildenv_info = get_xbuildenv_version_info(env, pyodide_version, pyodide_build_version)
    validate_pyodide_build_version(
        xbuildenv_info=xbuildenv_info,
        pyodide_build_version=pyodide_build_version,
    )

    emscripten_version = xbuildenv_info["emscripten"]
    log.step(f"Installing Emscripten version: {emscripten_version} ...")
    emcc_path = install_emscripten(tmp, emscripten_version)

    env["PATH"] = os.pathsep.join([str(emcc_path.parent), env["PATH"]])

    log.step(f"Installing Pyodide xbuildenv version: {pyodide_version} ...")
    env["PYODIDE_ROOT"] = install_xbuildenv(env, pyodide_build_version, pyodide_version)

    return env


def all_python_configurations() -> list[PythonConfiguration]:
    full_python_configs = resources.read_python_configs("pyodide")
    return [PythonConfiguration(**item) for item in full_python_configs]


def get_python_configurations(
    build_selector: BuildSelector,
    architectures: Set[Architecture],  # noqa: ARG001
) -> list[PythonConfiguration]:
    return [c for c in all_python_configurations() if build_selector(c.identifier)]


def build(options: Options, tmp_path: Path) -> None:
    python_configurations = get_python_configurations(
        options.globals.build_selector, options.globals.architectures
    )

    if not python_configurations:
        return

    try:
        before_all_options_identifier = python_configurations[0].identifier
        before_all_options = options.build_options(before_all_options_identifier)

        if before_all_options.before_all:
            log.step("Running before_all...")
            env = before_all_options.environment.as_dictionary(prev_environment=os.environ)
            before_all_prepared = prepare_command(
                before_all_options.before_all, project=".", package=before_all_options.package_dir
            )
            shell(before_all_prepared, env=env)

        built_wheels: list[Path] = []

        for config in python_configurations:
            build_options = options.build_options(config.identifier)
            build_frontend = build_options.build_frontend

            if build_frontend.name == "pip":
                msg = "The pyodide platform doesn't support pip frontend"
                raise errors.FatalError(msg)

            log.build_start(config.identifier)

            identifier_tmp_dir = tmp_path / config.identifier

            built_wheel_dir = identifier_tmp_dir / "built_wheel"
            repaired_wheel_dir = identifier_tmp_dir / "repaired_wheel"
            identifier_tmp_dir.mkdir()
            built_wheel_dir.mkdir()
            repaired_wheel_dir.mkdir()

            constraints_path = build_options.dependency_constraints.get_for_python_version(
                version=config.version, variant="pyodide", tmp_dir=identifier_tmp_dir
            )

            env = setup_python(
                tmp=identifier_tmp_dir / "build",
                python_configuration=config,
                constraints_path=constraints_path,
                environment=build_options.environment,
                user_pyodide_version=build_options.pyodide_version,
            )
            pip_version = get_pip_version(env)
            # The Pyodide command line runner mounts all directories in the host
            # filesystem into the Pyodide file system, except for the custom
            # file systems /dev, /lib, /proc, and /tmp. Mounting the mount
            # points for alternate file systems causes some mysterious failure
            # of the process (it just quits without any clear error).
            #
            # Because of this, by default Pyodide can't see anything under /tmp.
            # This environment variable tells it also to mount our temp
            # directory.
            oldmounts = ""
            extra_mounts = [str(identifier_tmp_dir)]
            if Path.cwd().is_relative_to("/tmp"):
                extra_mounts.append(str(Path.cwd()))

            if "_PYODIDE_EXTRA_MOUNTS" in env:
                oldmounts = env["_PYODIDE_EXTRA_MOUNTS"] + ":"
            env["_PYODIDE_EXTRA_MOUNTS"] = oldmounts + ":".join(extra_mounts)

            compatible_wheel = find_compatible_wheel(built_wheels, config.identifier)
            if compatible_wheel:
                log.step_end()
                print(
                    f"\nFound previously built wheel {compatible_wheel.name}, that's compatible with {config.identifier}. Skipping build step..."
                )
                built_wheel = compatible_wheel
            else:
                if build_options.before_build:
                    log.step("Running before_build...")
                    before_build_prepared = prepare_command(
                        build_options.before_build, project=".", package=build_options.package_dir
                    )
                    shell(before_build_prepared, env=env)

                log.step("Building wheel...")

                extra_flags = get_build_frontend_extra_flags(
                    build_frontend,
                    build_options.build_verbosity,
                    build_options.config_settings,
                    py38=False,
                )

                call(
                    "pyodide",
                    "build",
                    build_options.package_dir,
                    f"--outdir={built_wheel_dir}",
                    *extra_flags,
                    env=env,
                )
                built_wheel = next(built_wheel_dir.glob("*.whl"))

                if built_wheel.name.endswith("none-any.whl"):
                    raise errors.NonPlatformWheelError()

                if build_options.repair_command:
                    log.step("Repairing wheel...")

                    repair_command_prepared = prepare_command(
                        build_options.repair_command,
                        wheel=built_wheel,
                        dest_dir=repaired_wheel_dir,
                        package=build_options.package_dir,
                        project=".",
                    )
                    shell(repair_command_prepared, env=env)
                    log.step_end()
                else:
                    shutil.move(str(built_wheel), repaired_wheel_dir)

                repaired_wheel = next(repaired_wheel_dir.glob("*.whl"))

                if repaired_wheel.name in {wheel.name for wheel in built_wheels}:
                    raise errors.AlreadyBuiltWheelError(repaired_wheel.name)

            if build_options.test_command and build_options.test_selector(config.identifier):
                log.step("Testing wheel...")

                venv_dir = identifier_tmp_dir / "venv-test"
                # set up a virtual environment to install and test from, to make sure
                # there are no dependencies that were pulled in at build time.

                virtualenv_env = env.copy()
                virtualenv_env["PATH"] = os.pathsep.join(
                    [
                        str(ensure_node(config.node_version)),
                        virtualenv_env["PATH"],
                    ]
                )

                # pyodide venv uses virtualenv under the hood
                # use the pip embedded with virtualenv & disable network updates
                virtualenv_create_env = virtualenv_env.copy()
                virtualenv_create_env["VIRTUALENV_PIP"] = pip_version
                virtualenv_create_env["VIRTUALENV_NO_PERIODIC_UPDATE"] = "1"

                call("pyodide", "venv", venv_dir, env=virtualenv_create_env)

                virtualenv_env["PATH"] = os.pathsep.join(
                    [
                        str(venv_dir / "bin"),
                        virtualenv_env["PATH"],
                    ]
                )
                virtualenv_env["VIRTUAL_ENV"] = str(venv_dir)

                virtualenv_env = build_options.test_environment.as_dictionary(
                    prev_environment=virtualenv_env
                )

                # check that we are using the Python from the virtual environment
                call("which", "python", env=virtualenv_env)

                if build_options.before_test:
                    before_test_prepared = prepare_command(
                        build_options.before_test,
                        project=".",
                        package=build_options.package_dir,
                        wheel=repaired_wheel,
                    )
                    shell(before_test_prepared, env=virtualenv_env)

                # install the wheel
                call(
                    "pip",
                    "install",
                    f"{repaired_wheel}{build_options.test_extras}",
                    env=virtualenv_env,
                )

                # test the wheel
                if build_options.test_requires:
                    call("pip", "install", *build_options.test_requires, env=virtualenv_env)

                # run the tests from a temp dir, with an absolute path in the command
                # (this ensures that Python runs the tests against the installed wheel
                # and not the repo code)
                test_command_prepared = prepare_command(
                    build_options.test_command,
                    project=Path.cwd(),
                    package=build_options.package_dir.resolve(),
                )

                test_cwd = identifier_tmp_dir / "test_cwd"
                test_cwd.mkdir(exist_ok=True)

                if build_options.test_sources:
                    copy_test_sources(
                        build_options.test_sources,
                        Path.cwd(),
                        test_cwd,
                    )
                else:
                    # Use the test_fail.py file to raise a nice error if the user
                    # tries to run tests in the cwd
                    (test_cwd / "test_fail.py").write_text(resources.TEST_FAIL_CWD_FILE.read_text())

                shell(test_command_prepared, cwd=test_cwd, env=virtualenv_env)

            # we're all done here; move it to output (overwrite existing)
            output_wheel: Path | None = None
            if compatible_wheel is None:
                output_wheel = build_options.output_dir.joinpath(repaired_wheel.name)
                moved_wheel = move_file(repaired_wheel, output_wheel)
                if moved_wheel != output_wheel.resolve():
                    log.warning(
                        f"{repaired_wheel} was moved to {moved_wheel} instead of {output_wheel}"
                    )
                built_wheels.append(output_wheel)
            log.build_end(output_wheel)

    except subprocess.CalledProcessError as error:
        msg = f"Command {error.cmd} failed with code {error.returncode}. {error.stdout or ''}"
        raise errors.FatalError(msg) from error



import os
import subprocess
import sys
import textwrap
from pathlib import Path

import pytest

from . import test_projects, utils

project_with_environment_asserts = test_projects.new_c_project(
    setup_py_add=textwrap.dedent(
        r"""
        import os

        # explode if environment isn't correct, as set in CIBW_ENVIRONMENT
        CIBW_TEST_VAR = os.environ.get("CIBW_TEST_VAR")
        CIBW_TEST_VAR_2 = os.environ.get("CIBW_TEST_VAR_2")
        CIBW_TEST_VAR_3 = os.environ.get("CIBW_TEST_VAR_3")
        PATH = os.environ.get("PATH")

        if CIBW_TEST_VAR != "a b c":
            raise Exception('CIBW_TEST_VAR should equal "a b c". It was "%s"' % CIBW_TEST_VAR)
        if CIBW_TEST_VAR_2 != "1":
            raise Exception('CIBW_TEST_VAR_2 should equal "1". It was "%s"' % CIBW_TEST_VAR_2)
        if CIBW_TEST_VAR_3 != "test string 3":
            raise Exception('CIBW_TEST_VAR_3 should equal "test string 3". It was "%s"' % CIBW_TEST_VAR_3)
        if "/opt/cibw_test_path" not in PATH:
            raise Exception('PATH should contain "/opt/cibw_test_path". It was "%s"' % PATH)
        if "$PATH" in PATH:
            raise Exception('$PATH should be expanded in PATH. It was "%s"' % PATH)
        """
    )
)


def test(tmp_path: Path) -> None:
    python_echo = f"'{sys.executable}' -c \"import sys; print(*sys.argv[1:])\""
    project_dir = tmp_path / "project"
    project_with_environment_asserts.generate(project_dir)

    # write some information into the CIBW_ENVIRONMENT, for expansion and
    # insertion into the environment by cibuildwheel. This is checked
    # in setup.py
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_ENVIRONMENT": """CIBW_TEST_VAR="a b c" CIBW_TEST_VAR_2=1 CIBW_TEST_VAR_3="$(echo 'test string 3')" PATH=$PATH:/opt/cibw_test_path""",
            "CIBW_ENVIRONMENT_WINDOWS": f'''CIBW_TEST_VAR="a b c" CIBW_TEST_VAR_2=1 CIBW_TEST_VAR_3="$({python_echo} 'test string 3')" PATH="$PATH;/opt/cibw_test_path"''',
        },
        single_python=True,
    )

    # also check that we got the right wheels built
    expected_wheels = utils.expected_wheels("spam", "0.1.0", single_python=True)
    assert set(actual_wheels) == set(expected_wheels)


def test_overridden_path(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    project_dir = tmp_path / "project"
    output_dir = tmp_path / "output"

    project = test_projects.new_c_project()
    project.generate(project_dir)
    output_dir.mkdir()

    # mess up PATH, somehow
    if utils.get_platform() == "linux":
        with pytest.raises(subprocess.CalledProcessError):
            utils.cibuildwheel_run(
                project_dir,
                output_dir=output_dir,
                add_env={
                    "CIBW_BEFORE_ALL": "mkdir new_path && touch new_path/python && chmod +x new_path/python",
                    "CIBW_ENVIRONMENT": '''PATH="$(pwd)/new_path:$PATH"''',
                },
            )
    else:
        new_path = tmp_path / "another_bin"
        new_path.mkdir()
        (new_path / "python").touch(mode=0o777)

        with pytest.raises(subprocess.CalledProcessError):
            utils.cibuildwheel_run(
                project_dir,
                output_dir=output_dir,
                add_env={
                    "NEW_PATH": str(new_path),
                    "CIBW_ENVIRONMENT": f'''PATH="$NEW_PATH{os.pathsep}$PATH"''',
                },
            )

    assert len(list(output_dir.iterdir())) == 0
    captured = capfd.readouterr()
    assert "python available on PATH doesn't match our installed instance" in captured.err.replace(
        "venv", "installed"
    )


@pytest.mark.parametrize(
    "build_frontend",
    [
        pytest.param("pip", marks=utils.skip_if_pyodide("No pip for pyodide")),
        pytest.param(
            "build",
            marks=utils.skip_if_pyodide(
                "pyodide doesn't support multiple values for PIP_CONSTRAINT"
            ),
        ),
    ],
)
def test_overridden_pip_constraint(tmp_path: Path, build_frontend: str) -> None:
    """
    Verify that users can use PIP_CONSTRAINT to specify a specific version of
    a build-system.requires dependency, by asserting the version of pytz in the
    setup.py.
    """
    project_dir = tmp_path / "project"

    project = test_projects.new_c_project(
        setup_py_add=textwrap.dedent(
            """
            import pytz
            assert pytz.__version__ == "2022.4", f"{pytz.__version__!r} != '2022.4'"
        """
        )
    )
    project.files["pyproject.toml"] = textwrap.dedent(
        """
        [build-system]
        requires = ["setuptools", "pytz"]
        build-backend = "setuptools.build_meta"
    """
    )
    project.generate(project_dir)

    if utils.get_platform() == "linux":
        # put the constraints file in the project directory, so it's available
        # in the docker container
        constraints_file = project_dir / "constraints.txt"
    else:
        constraints_file = tmp_path / "constraints.txt"

    constraints_file.write_text("pytz==2022.4")

    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_BUILD_FRONTEND": build_frontend,
            "PIP_CONSTRAINT": str(constraints_file),
            "CIBW_ENVIRONMENT_LINUX": "PIP_CONSTRAINT=./constraints.txt",
        },
        single_python=True,
    )

    expected_wheels = utils.expected_wheels("spam", "0.1.0", single_python=True)
    assert set(actual_wheels) == set(expected_wheels)



import fnmatch
import functools
import json
import platform
import typing
from pathlib import Path

from filelock import FileLock

from cibuildwheel.util.file import download, extract_tar
from cibuildwheel.util.resources import PYTHON_BUILD_STANDALONE_RELEASES


class PythonBuildStandaloneAsset(typing.TypedDict):
    name: str
    url: str


class PythonBuildStandaloneRelease(typing.TypedDict):
    tag: str
    assets: list[PythonBuildStandaloneAsset]


class PythonBuildStandaloneReleaseData(typing.TypedDict):
    releases: list[PythonBuildStandaloneRelease]


@functools.cache
def get_python_build_standalone_release_data() -> PythonBuildStandaloneReleaseData:
    with open(PYTHON_BUILD_STANDALONE_RELEASES, "rb") as f:
        return typing.cast("PythonBuildStandaloneReleaseData", json.load(f))


class PythonBuildStandaloneError(Exception):
    """Errors related to python-build-standalone."""


def _get_platform_identifiers() -> tuple[str, str, str | None]:
    """
    Detects the current platform and returns architecture, platform, and libc
    identifiers.
    """
    system = platform.system()
    machine = platform.machine()
    machine_lower = machine.lower()

    arch_identifier: str
    platform_identifier: str
    libc_identifier: str | None = None

    # Map Architecture
    if machine_lower in ["x86_64", "amd64"]:
        arch_identifier = "x86_64"
    elif machine_lower in ["aarch64", "arm64"]:
        arch_identifier = "aarch64"
    else:
        msg = f"Unsupported architecture: {system} {machine}. Cannot download appropriate Python build."
        raise PythonBuildStandaloneError(msg)

    # Map OS + Libc
    if system == "Linux":
        platform_identifier = "unknown-linux"
        libc_identifier = "musl" if "musl" in (platform.libc_ver() or ("", "")) else "gnu"
    elif system == "Darwin":
        platform_identifier = "apple-darwin"
    elif system == "Windows":
        platform_identifier = "pc-windows-msvc"
    else:
        msg = f"Unsupported operating system: {system}. Cannot download appropriate Python build."
        raise PythonBuildStandaloneError(msg)

    print(
        f"Detected platform: arch='{arch_identifier}', platform='{platform_identifier}', libc='{libc_identifier}'"
    )
    return arch_identifier, platform_identifier, libc_identifier


def _get_pbs_asset(
    *,
    python_version: str,
    arch_identifier: str,
    platform_identifier: str,
    libc_identifier: str | None,
) -> tuple[str, str, str]:
    """Finds the asset, returning (tag, filename, url)."""
    release_data = get_python_build_standalone_release_data()

    expected_suffix = f"{arch_identifier}-{platform_identifier}"
    if libc_identifier:
        expected_suffix += f"-{libc_identifier}"
    expected_suffix += "-install_only.tar.gz"

    asset_pattern = f"cpython-{python_version}.*-{expected_suffix}"
    print(f"Looking for file with pattern {asset_pattern}")

    for release in release_data["releases"]:
        for asset in release["assets"]:
            asset_name = asset["name"]
            if not fnmatch.fnmatch(asset_name, asset_pattern):
                continue

            asset_url = asset["url"]
            return release["tag"], asset_url, asset_name

    # If loop completes without finding a match
    msg = f"Could not find python-build-standalone release asset matching {asset_pattern!r}."
    raise PythonBuildStandaloneError(msg)


def _download_or_get_from_cache(asset_url: str, asset_filename: str, cache_dir: Path) -> Path:
    with FileLock(cache_dir / (asset_filename + ".lock")):
        asset_cache_path = cache_dir / asset_filename
        if asset_cache_path.is_file():
            print(f"Using cached python_build_standalone: {asset_cache_path}")
            return asset_cache_path

        print(f"Downloading python_build_standalone: {asset_url} to {asset_cache_path}")
        download(asset_url, asset_cache_path)
        return asset_cache_path


def _find_python_executable(extracted_dir: Path) -> Path:
    """Finds the python executable within the extracted directory structure."""
    # Structure is typically 'python/bin/python' or 'python/python.exe'
    base_install_dir = extracted_dir / "python"

    if platform.system() == "Windows":
        executable_path = base_install_dir / "python.exe"
    else:
        executable_path = base_install_dir / "bin" / "python"

    if not executable_path.is_file():
        msg = f"Could not locate python executable at expected path {executable_path} within {extracted_dir}."
        raise PythonBuildStandaloneError(msg)

    print(f"Found python executable: {executable_path}")
    return executable_path.resolve()  # Return absolute path


def create_python_build_standalone_environment(
    python_version: str, temp_dir: Path, cache_dir: Path
) -> Path:
    """
    Returns a Python environment from python-build-standalone, downloading it
    if necessary using a cache, and expanding it into a fresh base path.

    Args:
        python_version: The Python version string (e.g., "3.12").
        temp_dir: A directory where the Python environment will be created.
        cache_dir: A directory to store/retrieve downloaded archives.

    Returns:
        The absolute path to the python executable within the created environment (in temp_dir).

    Raises:
        PythonBuildStandaloneError: If the platform is unsupported, the build cannot be found,
                    download/extraction fails, or configuration is invalid.
    """

    print(f"Creating python-build-standalone environment: version={python_version}")

    arch_id, platform_id, libc_id = _get_platform_identifiers()

    pbs_tag, asset_url, asset_filename = _get_pbs_asset(
        python_version=python_version,
        arch_identifier=arch_id,
        platform_identifier=platform_id,
        libc_identifier=libc_id,
    )

    print(f"Using python-build-standalone release: {pbs_tag}")

    archive_path = _download_or_get_from_cache(
        asset_url=asset_url, asset_filename=asset_filename, cache_dir=cache_dir
    )

    python_base_dir = temp_dir / "pbs"
    assert not python_base_dir.exists()
    extract_tar(archive_path, python_base_dir)

    return _find_python_executable(python_base_dir)



import os
import platform as platform_module
import textwrap
import unittest.mock
from collections.abc import Sequence
from pathlib import Path
from typing import Literal

import pytest

from cibuildwheel import errors
from cibuildwheel.bashlex_eval import local_environment_executor
from cibuildwheel.frontend import BuildFrontendConfig, get_build_frontend_extra_flags
from cibuildwheel.logger import Logger
from cibuildwheel.options import (
    CommandLineArguments,
    Options,
    _get_pinned_container_images,
)
from cibuildwheel.platforms import ALL_PLATFORM_MODULES, get_build_identifiers
from cibuildwheel.selector import EnableGroup
from cibuildwheel.util import resources
from cibuildwheel.util.packaging import DependencyConstraints

PYPROJECT_1 = """
[tool.cibuildwheel]
build = ["cp38-*", "cp313-*"]
skip = ["*musllinux*"]
archs = ["auto64", "auto32"]
environment = {FOO="BAR"}

test-command = "pyproject"
test-sources = ["test", "other dir"]

manylinux-x86_64-image = "manylinux_2_28"

environment-pass = ["EXAMPLE_ENV"]

pyodide-version = "0.28.0"

[tool.cibuildwheel.macos]
test-requires = "else"

[[tool.cibuildwheel.overrides]]
select = "cp313-*"
test-command = "pyproject-override"
manylinux-x86_64-image = "manylinux_2_34"
"""


def test_options_1(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    with tmp_path.joinpath("pyproject.toml").open("w") as f:
        f.write(PYPROJECT_1)

    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    monkeypatch.setattr(platform_module, "machine", lambda: "x86_64")

    options = Options(platform="linux", command_line_arguments=args, env={})

    module = ALL_PLATFORM_MODULES["linux"]
    identifiers = get_build_identifiers(
        platform_module=module,
        build_selector=options.globals.build_selector,
        architectures=options.globals.architectures,
    )

    override_display = """\
  *: pyproject
  cp313-manylinux_x86_64, cp313-manylinux_i686: pyproject-override"""
    print(options.summary(identifiers))

    assert override_display in options.summary(identifiers)

    default_build_options = options.build_options(identifier=None)

    assert default_build_options.environment.as_dictionary(prev_environment={}) == {"FOO": "BAR"}

    all_pinned_container_images = _get_pinned_container_images()
    pinned_x86_64_container_image = all_pinned_container_images["x86_64"]

    local = options.build_options("cp38-manylinux_x86_64")
    assert local.manylinux_images is not None
    assert local.test_command == "pyproject"
    assert local.test_sources == ["test", "other dir"]
    assert local.manylinux_images["x86_64"] == pinned_x86_64_container_image["manylinux_2_28"]

    local = options.build_options("cp313-manylinux_x86_64")
    assert local.manylinux_images is not None
    assert local.test_command == "pyproject-override"
    assert local.test_sources == ["test", "other dir"]
    assert local.manylinux_images["x86_64"] == pinned_x86_64_container_image["manylinux_2_34"]

    local = options.build_options("cp312-pyodide_wasm32")
    assert local.pyodide_version == "0.28.0"


def test_passthrough(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    with tmp_path.joinpath("pyproject.toml").open("w") as f:
        f.write(PYPROJECT_1)

    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    monkeypatch.setattr(platform_module, "machine", lambda: "x86_64")

    options = Options(platform="linux", command_line_arguments=args, env={"EXAMPLE_ENV": "ONE"})

    default_build_options = options.build_options(identifier=None)

    assert default_build_options.environment.as_dictionary(prev_environment={}) == {
        "FOO": "BAR",
        "EXAMPLE_ENV": "ONE",
    }


@pytest.mark.parametrize(
    "env_var_value",
    [
        "normal value",
        '"value wrapped in quotes"',
        "an unclosed single-quote: '",
        'an unclosed double-quote: "',
        "string\nwith\ncarriage\nreturns\n",
        "a trailing backslash \\",
    ],
)
def test_passthrough_evil(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, env_var_value: str
) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    monkeypatch.setattr(platform_module, "machine", lambda: "x86_64")
    options = Options(
        platform="linux",
        command_line_arguments=args,
        env={"CIBW_ENVIRONMENT_PASS_LINUX": "ENV_VAR", "ENV_VAR": env_var_value},
    )

    parsed_environment = options.build_options(identifier=None).environment
    assert parsed_environment.as_dictionary(prev_environment={}) == {"ENV_VAR": env_var_value}


xfail_env_parse = pytest.mark.xfail(
    raises=errors.ConfigurationError,
    reason="until we can figure out the right way to quote these values",
)


@pytest.mark.parametrize(
    "env_var_value",
    [
        "normal value",
        pytest.param('"value wrapped in quotes"', marks=[xfail_env_parse]),
        pytest.param('an unclosed double-quote: "', marks=[xfail_env_parse]),
        "string\nwith\ncarriage\nreturns\n",
        pytest.param("a trailing backslash \\", marks=[xfail_env_parse]),
    ],
)
def test_toml_environment_evil(tmp_path: Path, env_var_value: str) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    tmp_path.joinpath("pyproject.toml").write_text(
        textwrap.dedent(
            f"""\
            [tool.cibuildwheel.environment]
            EXAMPLE='''{env_var_value}'''
            """
        )
    )

    options = Options(platform="linux", command_line_arguments=args, env={})
    parsed_environment = options.build_options(identifier=None).environment
    assert parsed_environment.as_dictionary(prev_environment={}) == {"EXAMPLE": env_var_value}


@pytest.mark.parametrize(
    ("toml_assignment", "result_value"),
    [
        ('TEST_VAR="simple_value"', "simple_value"),
        # spaces
        ('TEST_VAR="simple value"', "simple value"),
        # env var
        ('TEST_VAR="$PARAM"', "spam"),
        ('TEST_VAR="$PARAM $PARAM"', "spam spam"),
        # env var extension
        ('TEST_VAR="before:$PARAM:after"', "before:spam:after"),
        # env var extension with spaces
        ('TEST_VAR="before $PARAM after"', "before spam after"),
        # literal $ - this test is just for reference, I'm not sure if this
        # syntax will work if we change the TOML quoting behaviour
        (r'TEST_VAR="before\\$after"', "before$after"),
    ],
)
def test_toml_environment_quoting(tmp_path: Path, toml_assignment: str, result_value: str) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    tmp_path.joinpath("pyproject.toml").write_text(
        textwrap.dedent(
            f"""\
            [tool.cibuildwheel.environment]
            {toml_assignment}
            """
        )
    )

    options = Options(platform="linux", command_line_arguments=args, env={})
    parsed_environment = options.build_options(identifier=None).environment
    environment_values = parsed_environment.as_dictionary(
        prev_environment={**os.environ, "PARAM": "spam"},
        executor=local_environment_executor,
    )

    assert environment_values["TEST_VAR"] == result_value


@pytest.mark.parametrize(
    ("toml_assignment", "result_name", "result_create_args", "result_disable_host_mount"),
    [
        (
            'container-engine = "podman"',
            "podman",
            (),
            False,
        ),
        (
            'container-engine = {name = "podman"}',
            "podman",
            (),
            False,
        ),
        (
            'container-engine = "docker; create_args: --some-option"',
            "docker",
            ("--some-option",),
            False,
        ),
        (
            'container-engine = {name = "docker", create-args = ["--some-option"]}',
            "docker",
            ("--some-option",),
            False,
        ),
        (
            'container-engine = {name = "docker", create-args = ["--some-option", "value that contains spaces"]}',
            "docker",
            ("--some-option", "value that contains spaces"),
            False,
        ),
        (
            'container-engine = {name = "docker", create-args = ["--some-option", "value;that;contains;semicolons"]}',
            "docker",
            ("--some-option", "value;that;contains;semicolons"),
            False,
        ),
        (
            'container-engine = {name = "docker", disable-host-mount = true}',
            "docker",
            (),
            True,
        ),
        (
            'container-engine = {name = "docker", disable_host_mount = true}',
            "docker",
            (),
            True,
        ),
    ],
)
def test_container_engine_option(
    tmp_path: Path,
    toml_assignment: str,
    result_name: str,
    result_create_args: tuple[str, ...],
    result_disable_host_mount: bool,
) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    tmp_path.joinpath("pyproject.toml").write_text(
        textwrap.dedent(
            f"""\
            [tool.cibuildwheel]
            {toml_assignment}
            """
        )
    )

    options = Options(platform="linux", command_line_arguments=args, env={})
    parsed_container_engine = options.build_options(None).container_engine

    assert parsed_container_engine.name == result_name
    assert parsed_container_engine.create_args == result_create_args
    assert parsed_container_engine.disable_host_mount == result_disable_host_mount


def test_environment_pass_references() -> None:
    options = Options(
        platform="linux",
        command_line_arguments=CommandLineArguments.defaults(),
        env={
            "CIBW_ENVIRONMENT_PASS_LINUX": "STARTER MAIN_COURSE",
            "STARTER": "green eggs",
            "MAIN_COURSE": "ham",
            "CIBW_ENVIRONMENT": 'MEAL="$STARTER and $MAIN_COURSE"',
        },
    )
    parsed_environment = options.build_options(identifier=None).environment
    assert parsed_environment.as_dictionary(prev_environment={}) == {
        "MEAL": "green eggs and ham",
        "STARTER": "green eggs",
        "MAIN_COURSE": "ham",
    }


@pytest.mark.parametrize(
    ("toml_assignment", "result_name", "result_args"),
    [
        (
            "",
            None,
            None,
        ),
        (
            'build-frontend = "build"',
            "build",
            [],
        ),
        (
            'build-frontend = "uv"',
            "uv",
            [],
        ),
        (
            'build-frontend = {name = "build"}',
            "build",
            [],
        ),
        (
            'build-frontend = "pip; args: --some-option"',
            "pip",
            ["--some-option"],
        ),
        (
            'build-frontend = {name = "pip", args = ["--some-option"]}',
            "pip",
            ["--some-option"],
        ),
    ],
)
def test_build_frontend_option(
    tmp_path: Path, toml_assignment: str, result_name: str, result_args: list[str]
) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    tmp_path.joinpath("pyproject.toml").write_text(
        textwrap.dedent(
            f"""\
            [tool.cibuildwheel]
            {toml_assignment}
            """
        )
    )

    options = Options(platform="linux", command_line_arguments=args, env={})
    parsed_build_frontend = options.build_options(identifier=None).build_frontend

    if toml_assignment:
        assert parsed_build_frontend.name == result_name
        assert parsed_build_frontend.args == result_args
    else:
        assert parsed_build_frontend.name == "build"
        assert parsed_build_frontend.args == ()


def test_override_inherit_environment(tmp_path: Path) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    pyproject_toml: Path = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            """\
            [tool.cibuildwheel]
            environment = {FOO="BAR", "HAM"="EGGS"}

            [[tool.cibuildwheel.overrides]]
            select = "cp37*"
            inherit.environment = "append"
            environment = {FOO="BAZ", "PYTHON"="MONTY"}
            """
        )
    )

    options = Options(platform="linux", command_line_arguments=args, env={})
    parsed_environment = options.build_options(identifier=None).environment
    assert parsed_environment.as_dictionary(prev_environment={}) == {
        "FOO": "BAR",
        "HAM": "EGGS",
    }

    assert options.build_options("cp37-manylinux_x86_64").environment.as_dictionary(
        prev_environment={}
    ) == {
        "FOO": "BAZ",
        "HAM": "EGGS",
        "PYTHON": "MONTY",
    }


def test_override_inherit_environment_with_references(tmp_path: Path) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    pyproject_toml: Path = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            """\
            [tool.cibuildwheel]
            environment = {PATH="/opt/bin:$PATH"}

            [[tool.cibuildwheel.overrides]]
            select = "cp37*"
            inherit.environment = "append"
            environment = {PATH="/opt/local/bin:$PATH"}
            """
        )
    )

    options = Options(platform="linux", command_line_arguments=args, env={"MONTY": "PYTHON"})
    parsed_environment = options.build_options(identifier=None).environment
    prev_environment = {"PATH": "/usr/bin:/bin"}
    assert parsed_environment.as_dictionary(prev_environment=prev_environment) == {
        "PATH": "/opt/bin:/usr/bin:/bin",
    }

    assert options.build_options("cp37-manylinux_x86_64").environment.as_dictionary(
        prev_environment=prev_environment
    ) == {
        "PATH": "/opt/local/bin:/opt/bin:/usr/bin:/bin",
    }


@pytest.mark.parametrize(
    ("toml_assignment", "env", "enable_args", "expected_result"),
    [
        ("", {}, [], False),
        ("enable = ['cpython-freethreading']", {}, [], True),
        ("enable = []", {}, [], False),
        ("", {}, ["cpython-freethreading"], True),
        ("", {}, ["cpython-freethreading", "pypy"], True),
        ("", {"CIBW_ENABLE": "pypy"}, [], False),
        ("", {"CIBW_ENABLE": "cpython-freethreading"}, [], True),
        ("enable = []", {"CIBW_ENABLE": "cpython-freethreading"}, [], True),
        ("enable = ['cpython-freethreading']", {"CIBW_ENABLE": "pypy"}, [], True),
        ("enable = ['cpython-freethreading']", {}, ["pypy"], True),
        ("enable = ['cpython-freethreading']", {"CIBW_ENABLE": ""}, [], True),
        ("enable = []", {"CIBW_ENABLE": ""}, [], False),
    ],
)
def test_free_threaded_support(
    tmp_path: Path,
    toml_assignment: str,
    env: dict[str, str],
    enable_args: list[str],
    expected_result: bool,
) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path
    args.enable = enable_args

    pyproject_toml: Path = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            f"""\
            [tool.cibuildwheel]
            {toml_assignment}
            """
        )
    )
    options = Options(platform="linux", command_line_arguments=args, env=env)
    if expected_result:
        assert EnableGroup.CPythonFreeThreading in options.globals.build_selector.enable
    else:
        assert EnableGroup.CPythonFreeThreading not in options.globals.build_selector.enable


@pytest.mark.parametrize(
    ("toml_assignment", "base_file_path", "packages"),
    [
        ("", resources.CONSTRAINTS, []),
        ("dependency-versions = 'pinned'", resources.CONSTRAINTS, []),
        ("dependency-versions = 'latest'", None, []),
        ("dependency-versions = 'constraints file.txt'", Path("constraints file.txt"), []),
        (
            "dependency-versions = \"file:'constraints file.txt'\"",
            Path("constraints file.txt"),
            [],
        ),
        (
            "dependency-versions = {file = 'constraints file.txt'}",
            Path("constraints file.txt"),
            [],
        ),
        (
            "dependency-versions = 'packages: foo==1.2.3 bar==4.5.6'",
            None,
            ["foo==1.2.3", "bar==4.5.6"],
        ),
    ],
)
def test_dependency_versions_toml(
    tmp_path: Path,
    toml_assignment: str,
    base_file_path: Path | None,
    packages: list[str] | None,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    (tmp_path / "constraints file.txt").write_text("")
    monkeypatch.chdir(tmp_path)

    pyproject_toml: Path = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            f"""\
            [tool.cibuildwheel]
            {toml_assignment}
            """
        )
    )

    options = Options(platform="linux", command_line_arguments=args, env={})
    parsed_dependency_constraints = options.build_options(None).dependency_constraints
    if base_file_path is None and packages is None:
        assert parsed_dependency_constraints == DependencyConstraints.latest()
    else:
        if parsed_dependency_constraints.base_file_path and base_file_path:
            assert parsed_dependency_constraints.base_file_path.samefile(base_file_path)
        else:
            assert parsed_dependency_constraints.base_file_path == base_file_path
        assert parsed_dependency_constraints.packages == packages


@pytest.mark.parametrize(
    ("image", "deprecated", "raises"),
    [
        ("manylinux1", True, True),
        ("manylinux2010", True, True),
        ("manylinux2014", False, False),
        ("manylinux_2_24", True, True),
        ("manylinux_2_28", False, False),
        ("manylinux_2_34", False, False),
        ("musllinux_1_1", True, True),
        ("musllinux_1_2", False, False),
    ],
)
def test_deprecated_image(
    image: str, deprecated: bool, raises: bool, capsys: pytest.CaptureFixture[str]
) -> None:
    assert not raises or deprecated
    args = CommandLineArguments.defaults()
    env = {
        "CIBW_ARCHS": "x86_64",
        "CIBW_MANYLINUX_X86_64_IMAGE": image if image.startswith("manylinux") else "",
        "CIBW_MUSLLINUX_X86_64_IMAGE": image if image.startswith("musllinux") else "",
    }
    options = Options(platform="linux", command_line_arguments=args, env=env)
    try:
        bo = options.build_options(None)
        assert not raises
    except errors.DeprecationError:
        assert raises
        return
    images = bo.manylinux_images if image.startswith("manylinux") else bo.musllinux_images
    assert images is not None
    resolved_image = images["x86_64"]
    captured = capsys.readouterr()
    if deprecated:
        assert f"Deprecated image {image!r}" in captured.err
        assert f"{resolved_image!r}" in captured.err
    else:
        assert "Deprecated image" not in captured.err


@pytest.mark.parametrize(
    ("frontend", "verbosity", "result"),
    [
        ("pip", 3, ["-Ca", "-Cb", "-1", "-vvv"]),
        ("pip", 2, ["-Ca", "-Cb", "-1", "-vv"]),
        ("pip", -1, ["-Ca", "-Cb", "-1", "-q"]),
        ("uv", -1, ["-Ca", "-Cb", "-1", "-q"]),
        ("build", -1, ["-Ca", "-Cb", "-1", "-q"]),
        ("build[uv]", -1, ["-Ca", "-Cb", "-1", "-q"]),
        ("build", 0, ["-Ca", "-Cb", "-1"]),
        ("build", 1, ["-Ca", "-Cb", "-1"]),
        ("build", 2, ["-Ca", "-Cb", "-1", "-v"]),
        ("build", 3, ["-Ca", "-Cb", "-1", "-vv"]),
        ("build[uv]", 3, ["-Ca", "-Cb", "-1", "-vv"]),
        ("uv", 3, ["-Ca", "-Cb", "-1", "-vv"]),
    ],
)
def test_get_build_frontend_extra_flags(
    frontend: Literal["pip", "build", "build[uv]"],
    verbosity: int,
    result: list[str],
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    mock_warning = unittest.mock.MagicMock()
    monkeypatch.setattr(Logger, "warning", mock_warning)
    build_frontend = BuildFrontendConfig(frontend, ["-1"])
    args = get_build_frontend_extra_flags(
        build_frontend=build_frontend, verbosity_level=verbosity, config_settings="a b", py38=False
    )

    assert args == result
    mock_warning.assert_not_called()


@pytest.mark.parametrize("frontend", ["build", "build[uv]"])
def test_get_build_frontend_extra_flags_warning(
    frontend: Literal["build", "build[uv]"], monkeypatch: pytest.MonkeyPatch
) -> None:
    mock_warning = unittest.mock.MagicMock()
    monkeypatch.setattr(Logger, "warning", mock_warning)
    build_frontend = BuildFrontendConfig(frontend, ["-1"])
    args = get_build_frontend_extra_flags(
        build_frontend=build_frontend, verbosity_level=-1, config_settings="a b", py38=True
    )
    assert args == ["-Ca", "-Cb", "-1"]
    mock_warning.assert_called_once()


@pytest.mark.parametrize(
    ("definition", "expected_args"),
    [
        ("", ()),
        ('test-runtime = ""', ()),
        ("test-runtime = {}", ()),
        ('test-runtime = {args = ""}', []),
        ('test-runtime = "args: --simulator foo"', ["--simulator", "foo"]),
        ('test-runtime = {args = ["--simulator", "foo"]}', ["--simulator", "foo"]),
    ],
)
def test_test_runtime_handling(
    tmp_path: Path, definition: str, expected_args: Sequence[str] | None
) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    pyproject_toml: Path = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            f"""\
            [tool.cibuildwheel]
            {definition}
            """
        )
    )

    options = Options(platform="ios", command_line_arguments=args, env={})

    local = options.build_options("cp313-ios_13_0_arm64_iphoneos")
    assert local.test_runtime.args == expected_args


@pytest.mark.parametrize(
    ("definition", "expected"),
    [
        ("", None),
        ("xbuild-tools = []", []),
        ('xbuild-tools = ["cmake", "rustc"]', ["cmake", "rustc"]),
    ],
)
def test_xbuild_tools_handling(tmp_path: Path, definition: str, expected: list[str] | None) -> None:
    args = CommandLineArguments.defaults()
    args.package_dir = tmp_path

    pyproject_toml: Path = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            f"""\
            [tool.cibuildwheel]
            {definition}
            """
        )
    )

    options = Options(platform="ios", command_line_arguments=args, env={})

    local = options.build_options("cp313-ios_13_0_arm64_iphoneos")
    assert local.xbuild_tools == expected



import subprocess
from pathlib import Path

import pytest

from . import test_projects, utils


@pytest.mark.parametrize(
    "frontend_name",
    [
        pytest.param("pip", marks=utils.skip_if_pyodide("No pip for pyodide")),
        "build",
    ],
)
def test_build_frontend_args(
    tmp_path: Path, capfd: pytest.CaptureFixture[str], frontend_name: str
) -> None:
    project = test_projects.new_c_project()
    project_dir = tmp_path / "project"
    project.generate(project_dir)

    # the build will fail because the frontend is called with '-h' - it prints the help message
    add_env = {"CIBW_BUILD_FRONTEND": f"{frontend_name}; args: -h"}
    if utils.get_platform() == "pyodide":
        add_env["TERM"] = "dumb"  # disable color / style
        add_env["NO_COLOR"] = "1"
    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(project_dir, add_env=add_env, single_python=True)

    captured = capfd.readouterr()
    print(captured.out)

    # check that the help message was printed
    if frontend_name == "pip":
        assert "Usage:" in captured.out
        assert "Wheel Options:" in captured.out
    elif utils.get_platform() == "pyodide":
        assert "Usage: pyodide build" in captured.out
    else:
        assert "usage:" in captured.out
        assert "A simple, correct Python build frontend." in captured.out



#!/usr/bin/env python3

import os
import shutil
import subprocess
import sys
from pathlib import Path


def main() -> None:
    # CIBW_HOST_TRIPLET is set in the android_env to the Android target triplet.
    target = os.environ.get("CIBW_HOST_TRIPLET")

    cmd_name = Path(sys.argv[0]).name

    # Find the real command in PATH, excluding the current script's directory
    path_env = os.environ.get("PATH", "")
    script_dir = Path(__file__).resolve().parent

    paths = path_env.split(os.pathsep)
    # Filter out the script directory to avoid recursion
    filtered_paths = [p for p in paths if Path(p).resolve() != script_dir]
    filtered_path_env = os.pathsep.join(filtered_paths)

    real_cmd = shutil.which(cmd_name, path=filtered_path_env)

    if not real_cmd:
        sys.stderr.write(f"cibuildwheel: Error: Could not find system {cmd_name}\n")
        sys.exit(1)

    # If we have a target (i.e. we are in the android_env), try to install it.
    if target:
        # Check if rustup is available to install the target
        rustup_path = shutil.which("rustup", path=filtered_path_env)

        if rustup_path:
            try:
                # We call rustup to ensure the target is installed.
                subprocess.run(
                    [rustup_path, "target", "add", target],
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.PIPE,
                    text=True,
                )
            except subprocess.CalledProcessError as e:
                sys.stderr.write(
                    f"cibuildwheel: Error: Failed to install Rust target {target}: {e.stderr}\n"
                )
                sys.exit(1)

    # Execute the real command
    os.execv(real_cmd, [real_cmd, *sys.argv[1:]])


if __name__ == "__main__":
    main()



import contextlib
import json
import os
import random
import shutil
import subprocess
import sys
import textwrap
import time
from collections.abc import Iterator
from contextlib import nullcontext
from pathlib import Path, PurePath, PurePosixPath

import pytest
import tomli_w

import cibuildwheel.oci_container
from cibuildwheel.ci import CIProvider, detect_ci_provider
from cibuildwheel.environment import EnvironmentAssignmentBash
from cibuildwheel.errors import OCIEngineTooOldError
from cibuildwheel.oci_container import (
    OCIContainer,
    OCIContainerEngineConfig,
    OCIPlatform,
    _check_engine_version,
)

# Test utilities

# for these tests we use manylinux2014 images, because they're available on
# multi architectures and include python3.8
DEFAULT_IMAGE = "quay.io/pypa/manylinux2014:2025.03.08-1"
DEFAULT_OCI_PLATFORM = OCIPlatform.native()

PODMAN = OCIContainerEngineConfig(name="podman")


@pytest.fixture(params=["docker", "podman"], scope="module")
def container_engine(request: pytest.FixtureRequest) -> Iterator[OCIContainerEngineConfig]:
    if request.param == "docker" and not request.config.getoption("--run-docker"):
        pytest.skip("need --run-docker option to run")
    if request.param == "podman" and not request.config.getoption("--run-podman"):
        pytest.skip("need --run-podman option to run")

    def get_images() -> set[str]:
        if detect_ci_provider() is None:
            return set()
        images = subprocess.run(
            [request.param, "image", "ls", "--format", "{{json .ID}}"],
            text=True,
            check=True,
            stdout=subprocess.PIPE,
        ).stdout
        return {json.loads(image.strip()) for image in images.splitlines() if image.strip()}

    images_before = get_images()
    try:
        yield OCIContainerEngineConfig(name=request.param)
    finally:
        images_after = get_images()
        for image in images_after - images_before:
            subprocess.run([request.param, "rmi", image], check=False)


# Tests


def test_simple(container_engine: OCIContainerEngineConfig) -> None:
    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        assert container.call(["echo", "hello"], capture_output=True) == "hello\n"


def test_no_lf(container_engine: OCIContainerEngineConfig) -> None:
    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        assert container.call(["printf", "hello"], capture_output=True) == "hello"


def test_debug_info(container_engine: OCIContainerEngineConfig) -> None:
    container = OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    )
    print(container.debug_info())
    with container:
        pass


def test_environment(container_engine: OCIContainerEngineConfig) -> None:
    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        assert (
            container.call(
                ["sh", "-c", "echo $TEST_VAR"], env={"TEST_VAR": "1"}, capture_output=True
            )
            == "1\n"
        )


def test_environment_pass(
    container_engine: OCIContainerEngineConfig, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("CIBUILDWHEEL", "1")
    monkeypatch.setenv("SOURCE_DATE_EPOCH", "1489957071")
    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        assert container.call(["sh", "-c", "echo $CIBUILDWHEEL"], capture_output=True) == "1\n"
        assert (
            container.call(["sh", "-c", "echo $SOURCE_DATE_EPOCH"], capture_output=True)
            == "1489957071\n"
        )


def test_cwd(container_engine: OCIContainerEngineConfig) -> None:
    with OCIContainer(
        engine=container_engine,
        image=DEFAULT_IMAGE,
        oci_platform=DEFAULT_OCI_PLATFORM,
        cwd="/cibuildwheel/working_directory",
    ) as container:
        assert container.call(["pwd"], capture_output=True) == "/cibuildwheel/working_directory\n"
        assert container.call(["pwd"], capture_output=True, cwd="/opt") == "/opt\n"


def test_container_removed(container_engine: OCIContainerEngineConfig) -> None:
    # test is flaky on some platforms, implement retry for 5 second
    timeout = 50  # * 100 ms = 5s
    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        assert container.name is not None
        container_name = container.name
        for _ in range(timeout):
            docker_containers_listing = subprocess.run(
                f"{container.engine.name} container ls",
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
                text=True,
            ).stdout
            if container_name in docker_containers_listing:
                break
            time.sleep(0.1)
        assert container_name in docker_containers_listing

    for _ in range(timeout):
        docker_containers_listing = subprocess.run(
            f"{container.engine.name} container ls",
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            text=True,
        ).stdout
        if container_name not in docker_containers_listing:
            break
        time.sleep(0.1)
    assert container_name not in docker_containers_listing


def test_large_environment(container_engine: OCIContainerEngineConfig) -> None:
    # max environment variable size is 128kB
    long_env_var_length = 127 * 1024
    large_environment = {
        "a": "0" * long_env_var_length,
        "b": "0" * long_env_var_length,
        "c": "0" * long_env_var_length,
        "d": "0" * long_env_var_length,
    }

    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        # check the length of d
        assert (
            container.call(["sh", "-c", "echo ${#d}"], env=large_environment, capture_output=True)
            == f"{long_env_var_length}\n"
        )


def test_binary_output(container_engine: OCIContainerEngineConfig) -> None:
    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        # note: the below embedded snippets are in python2

        # check that we can pass though arbitrary binary data without erroring
        container.call(
            [
                "/usr/bin/python2",
                "-c",
                textwrap.dedent(
                    """
                    import sys
                    sys.stdout.write(''.join(chr(n) for n in range(0, 256)))
                    """
                ),
            ]
        )

        # check that we can capture arbitrary binary data
        output = container.call(
            [
                "/usr/bin/python2",
                "-c",
                textwrap.dedent(
                    """
                    import sys
                    sys.stdout.write(''.join(chr(n % 256) for n in range(0, 512)))
                    """
                ),
            ],
            capture_output=True,
        )

        data = bytes(output, encoding="utf8", errors="surrogateescape")

        for i in range(512):
            assert data[i] == i % 256

        # check that environment variables can carry binary data, except null characters
        # (https://www.gnu.org/software/libc/manual/html_node/Environment-Variables.html)
        binary_data = bytes(n for n in range(1, 256))
        binary_data_string = str(binary_data, encoding="utf8", errors="surrogateescape")
        output = container.call(
            ["python2", "-c", 'import os, sys; sys.stdout.write(os.environ["TEST_VAR"])'],
            env={"TEST_VAR": binary_data_string},
            capture_output=True,
        )
        assert output == binary_data_string


@pytest.mark.parametrize(
    "file_path",
    ["test.dat", "path/to/test.dat"],
)
def test_file_operation(
    tmp_path: Path, container_engine: OCIContainerEngineConfig, file_path: str
) -> None:
    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        # test copying a file in
        test_binary_data = bytes(random.randrange(256) for _ in range(1000))
        original_test_file = tmp_path / file_path
        original_test_file.parent.mkdir(parents=True, exist_ok=True)
        original_test_file.write_bytes(test_binary_data)

        dst_file = PurePath("/tmp") / file_path

        container.copy_into(original_test_file, dst_file)

        owner = container.call(["stat", "-c", "%u:%g", dst_file], capture_output=True).strip()
        assert owner == "0:0"

        output = container.call(["cat", dst_file], capture_output=True)
        assert test_binary_data == bytes(output, encoding="utf8", errors="surrogateescape")


def test_dir_operations(tmp_path: Path, container_engine: OCIContainerEngineConfig) -> None:
    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        test_binary_data = bytes(random.randrange(256) for _ in range(1000))
        original_test_file = tmp_path / "test.dat"
        original_test_file.write_bytes(test_binary_data)

        # test copying a dir in
        test_dir = tmp_path / "test_dir"
        test_dir.mkdir()
        test_file = test_dir / "test.dat"
        shutil.copyfile(original_test_file, test_file)

        dst_dir = PurePosixPath("/tmp/test_dir")
        dst_file = dst_dir / "test.dat"
        container.copy_into(test_dir, dst_dir)

        owner = container.call(["stat", "-c", "%u:%g", dst_dir], capture_output=True).strip()
        assert owner == "0:0"

        owner = container.call(["stat", "-c", "%u:%g", dst_file], capture_output=True).strip()
        assert owner == "0:0"

        output = container.call(["cat", dst_file], capture_output=True)
        assert test_binary_data == bytes(output, encoding="utf8", errors="surrogateescape")

        # test glob
        assert container.glob(dst_dir, "*.dat") == [dst_file]

        # test copy dir out
        new_test_dir = tmp_path / "test_dir_new"
        container.copy_out(dst_dir, new_test_dir)

        assert os.getuid() == new_test_dir.stat().st_uid
        assert os.getgid() == new_test_dir.stat().st_gid
        assert os.getuid() == (new_test_dir / "test.dat").stat().st_uid
        assert os.getgid() == (new_test_dir / "test.dat").stat().st_gid

        assert test_binary_data == (new_test_dir / "test.dat").read_bytes()


def test_environment_executor(container_engine: OCIContainerEngineConfig) -> None:
    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        assignment = EnvironmentAssignmentBash("TEST=$(echo 42)")
        assert assignment.evaluated_value({}, container.environment_executor) == "42"


def test_podman_vfs(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, container_engine: OCIContainerEngineConfig
) -> None:
    if container_engine.name != "podman":
        pytest.skip("only runs with podman")
    if sys.platform.startswith("darwin"):
        pytest.skip("Skipping test because podman on this platform does not support vfs")

    # create the VFS configuration
    vfs_path = tmp_path / "podman_vfs"
    vfs_path.mkdir()

    # This requires that we write configuration files and point to them
    # with environment variables before we run podman
    # https://github.com/containers/common/blob/main/docs/containers.conf.5.md
    vfs_containers_conf_data = {
        "containers": {
            "default_capabilities": [
                "CHOWN",
                "DAC_OVERRIDE",
                "FOWNER",
                "FSETID",
                "KILL",
                "NET_BIND_SERVICE",
                "SETFCAP",
                "SETGID",
                "SETPCAP",
                "SETUID",
                "SYS_CHROOT",
            ]
        },
        "engine": {"cgroup_manager": "cgroupfs", "events_logger": "file"},
    }
    # https://github.com/containers/storage/blob/main/docs/containers-storage.conf.5.md
    storage_root = vfs_path / ".local/share/containers/vfs-storage"
    run_root = vfs_path / ".local/share/containers/vfs-runroot"
    storage_root.mkdir(parents=True, exist_ok=True)
    run_root.mkdir(parents=True, exist_ok=True)
    vfs_containers_storage_conf_data = {
        "storage": {
            "driver": "vfs",
            "graphroot": os.fspath(storage_root),
            "runroot": os.fspath(run_root),
            "rootless_storage_path": os.fspath(storage_root),
            "options": {
                # "remap-user": "containers",
                "aufs": {"mountopt": "rw"},
                "overlay": {"mountopt": "rw", "force_mask": "shared"},
                # "vfs": {"ignore_chown_errors": "true"},
            },
        }
    }

    vfs_containers_conf_fpath = vfs_path / "temp_vfs_containers.conf"
    vfs_containers_storage_conf_fpath = vfs_path / "temp_vfs_containers_storage.conf"
    with open(vfs_containers_conf_fpath, "wb") as file:
        tomli_w.dump(vfs_containers_conf_data, file)

    with open(vfs_containers_storage_conf_fpath, "wb") as file:
        tomli_w.dump(vfs_containers_storage_conf_data, file)

    monkeypatch.setenv("CONTAINERS_CONF", str(vfs_containers_conf_fpath))
    monkeypatch.setenv("CONTAINERS_STORAGE_CONF", str(vfs_containers_storage_conf_fpath))

    with OCIContainer(
        engine=PODMAN, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        # test running a command
        assert container.call(["echo", "hello"], capture_output=True) == "hello\n"

        # test copying a file into the container
        (tmp_path / "some_file.txt").write_text("1234")
        container.copy_into(tmp_path / "some_file.txt", PurePosixPath("some_file.txt"))
        assert container.call(["cat", "some_file.txt"], capture_output=True) == "1234"

    # Clean up

    # When using the VFS, user is not given write permissions by default in
    # new directories. As a workaround we use 'podman unshare' to delete them
    # as UID 0. The reason why permission errors occur on podman is documented
    # in https://podman.io/blogs/2018/10/03/podman-remove-content-homedir.html
    subprocess.run(["podman", "unshare", "rm", "-rf", vfs_path], check=True)


def test_create_args_volume(tmp_path: Path, container_engine: OCIContainerEngineConfig) -> None:
    if container_engine.name != "docker":
        pytest.skip("only runs with docker")

    if "CIRCLECI" in os.environ or "GITLAB_CI" in os.environ:
        pytest.skip(
            "Skipping test on CircleCI/GitLab because docker there does not support --volume"
        )

    test_mount_dir = tmp_path / "test_mount"
    test_mount_dir.mkdir()
    (test_mount_dir / "test_file.txt").write_text("1234")
    container_engine = OCIContainerEngineConfig(
        name="docker", create_args=(f"--volume={test_mount_dir}:/test_mount",)
    )

    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        assert container.call(["cat", "/test_mount/test_file.txt"], capture_output=True) == "1234"


@pytest.mark.parametrize(
    ("config", "name", "create_args"),
    [
        (
            "docker",
            "docker",
            (),
        ),
        (
            "docker;create_args:",
            "docker",
            (),
        ),
        (
            "docker;create_args:--abc --def",
            "docker",
            ("--abc", "--def"),
        ),
        (
            "docker; create_args: --abc --def",
            "docker",
            ("--abc", "--def"),
        ),
        (
            "name:docker; create_args: --abc --def",
            "docker",
            ("--abc", "--def"),
        ),
        (
            'docker; create_args: --some-option="value with spaces"',
            "docker",
            ("--some-option=value with spaces",),
        ),
        (
            'docker; create_args: --some-option="value; with; semicolons" --another-option',
            "docker",
            ("--some-option=value; with; semicolons", "--another-option"),
        ),
        (
            "docker; create_args: --platform=linux/amd64",
            "docker",
            (),
        ),
        (
            "podman; create_args: --platform=linux/amd64",
            "podman",
            (),
        ),
        (
            "docker; create_args: --platform linux/amd64",
            "docker",
            (),
        ),
        (
            "podman; create_args: --platform linux/amd64",
            "podman",
            (),
        ),
    ],
)
def test_parse_engine_config(
    config: str, name: str, create_args: tuple[str, ...], capsys: pytest.CaptureFixture[str]
) -> None:
    engine_config = OCIContainerEngineConfig.from_config_string(config)
    assert engine_config.name == name
    assert engine_config.create_args == create_args
    if "--platform" in config:
        captured = capsys.readouterr()
        assert (
            "Using '--platform' in 'container-engine::create_args' is deprecated. It will be ignored."
            in captured.err
        )


@pytest.mark.skipif(DEFAULT_OCI_PLATFORM != OCIPlatform.AMD64, reason="Only runs on x86_64")
def test_enforce_32_bit(container_engine: OCIContainerEngineConfig) -> None:
    with OCIContainer(
        engine=container_engine, image=DEFAULT_IMAGE, oci_platform=OCIPlatform.i386
    ) as container:
        assert container.call(["uname", "-m"], capture_output=True).strip() == "i686"
        container_args = subprocess.run(
            f"{container.engine.name} inspect -f '{{{{json .Args }}}}' {container.name}",
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            text=True,
        ).stdout
        assert json.loads(container_args) == ["/bin/bash"]


@pytest.mark.parametrize(
    ("config", "should_have_host_mount"),
    [
        ("{name}", True),
        ("{name}; disable_host_mount: false", True),
        ("{name}; disable_host_mount: true", False),
    ],
)
def test_disable_host_mount(
    tmp_path: Path,
    container_engine: OCIContainerEngineConfig,
    config: str,
    should_have_host_mount: bool,
) -> None:
    if detect_ci_provider() in {CIProvider.circle_ci, CIProvider.gitlab}:
        pytest.skip("Skipping test because docker on this platform does not support host mounts")
    if sys.platform.startswith("darwin"):
        pytest.skip("Skipping test because docker on this platform does not support host mounts")

    engine = OCIContainerEngineConfig.from_config_string(config.format(name=container_engine.name))

    sentinel_file = tmp_path / "sentinel"
    sentinel_file.write_text("12345")

    with OCIContainer(
        engine=engine, image=DEFAULT_IMAGE, oci_platform=DEFAULT_OCI_PLATFORM
    ) as container:
        host_mount_path = "/host" + str(sentinel_file)
        if should_have_host_mount:
            assert container.call(["cat", host_mount_path], capture_output=True) == "12345"
        else:
            with pytest.raises(subprocess.CalledProcessError):
                container.call(["cat", host_mount_path], capture_output=True)


@pytest.mark.parametrize("platform", list(OCIPlatform))
def test_local_image(
    container_engine: OCIContainerEngineConfig, platform: OCIPlatform, tmp_path: Path
) -> None:
    if detect_ci_provider() == CIProvider.travis_ci and DEFAULT_OCI_PLATFORM not in {
        OCIPlatform.AMD64,
        platform,
    }:
        pytest.skip("Skipping test because docker on this platform does not support QEMU")
    if container_engine.name == "podman" and platform == OCIPlatform.ARMV7:
        # both GHA & local macOS arm64 podman desktop are failing
        pytest.xfail("podman fails with armv7l images")

    remote_image = "debian:trixie-slim"
    platform_name = platform.value.replace("/", "_")
    local_image = f"cibw_{container_engine.name}_{platform_name}_local:latest"
    dockerfile = tmp_path / "Dockerfile"
    dockerfile.write_text(f"FROM {remote_image}")
    subprocess.run(
        [container_engine.name, "pull", f"--platform={platform.value}", remote_image],
        check=True,
    )
    subprocess.run(
        [container_engine.name, "build", f"--platform={platform.value}", "-t", local_image, "."],
        check=True,
        cwd=tmp_path,
    )
    with OCIContainer(engine=container_engine, image=local_image, oci_platform=platform):
        pass


@pytest.mark.parametrize("platform", list(OCIPlatform))
def test_multiarch_image(container_engine: OCIContainerEngineConfig, platform: OCIPlatform) -> None:
    if detect_ci_provider() == CIProvider.travis_ci and DEFAULT_OCI_PLATFORM not in {
        OCIPlatform.AMD64,
        platform,
    }:
        pytest.skip("Skipping test because docker on this platform does not support QEMU")
    if container_engine.name == "podman" and platform == OCIPlatform.ARMV7:
        # both GHA & local macOS arm64 podman desktop are failing
        pytest.xfail("podman fails with armv7l images")
    with OCIContainer(
        engine=container_engine, image="debian:trixie-slim", oci_platform=platform
    ) as container:
        output = container.call(["uname", "-m"], capture_output=True)
        output_map_kernel = {
            OCIPlatform.i386: ("i686",),
            OCIPlatform.AMD64: ("x86_64",),
            OCIPlatform.ARMV7: ("armv7l", "armv8l"),
            OCIPlatform.ARM64: ("aarch64",),
            OCIPlatform.PPC64LE: ("ppc64le",),
            OCIPlatform.RISCV64: ("riscv64",),
            OCIPlatform.S390X: ("s390x",),
        }
        assert output.strip() in output_map_kernel[platform]
        output = container.call(["dpkg", "--print-architecture"], capture_output=True)
        output_map_dpkg = {
            OCIPlatform.i386: "i386",
            OCIPlatform.AMD64: "amd64",
            OCIPlatform.ARMV7: "armhf",
            OCIPlatform.ARM64: "arm64",
            OCIPlatform.PPC64LE: "ppc64el",
            OCIPlatform.RISCV64: "riscv64",
            OCIPlatform.S390X: "s390x",
        }
        assert output_map_dpkg[platform] == output.strip()


@pytest.mark.parametrize(
    ("engine_name", "version", "context"),
    [
        (
            "docker",
            None,  # 17.12.1-ce does supports "docker version --format '{{json . }}'" so a version before that
            pytest.raises(OCIEngineTooOldError),
        ),
        (
            "docker",
            '{"Client":{"Version":"19.03.15","ApiVersion": "1.40"},"Server":{"ApiVersion": "1.40"}}',
            pytest.raises(OCIEngineTooOldError),
        ),
        (
            "docker",
            '{"Client":{"Version":"20.10.0","ApiVersion":"1.41"},"Server":{"ApiVersion":"1.41"}}',
            nullcontext(),
        ),
        (
            "docker",
            '{"Client":{"Version":"24.0.0","ApiVersion":"1.43"},"Server":{"ApiVersion":"1.43"}}',
            nullcontext(),
        ),
        (
            "docker",
            '{"Client":{"Version":"29.0.0","ApiVersion":"1.52"},"Server":{"APIVersion":"1.52"}}',
            nullcontext(),
        ),
        (
            "docker",
            '{"Client":{"ApiVersion":"1.43"},"Server":{"ApiVersion":"1.30"}}',
            pytest.raises(OCIEngineTooOldError),
        ),
        (
            "docker",
            '{"Client":{"ApiVersion":"1.30"},"Server":{"ApiVersion":"1.43"}}',
            pytest.raises(OCIEngineTooOldError),
        ),
        ("podman", '{"Client":{"Version":"5.2.0"},"Server":{"Version":"5.1.2"}}', nullcontext()),
        ("podman", '{"Client":{"Version":"4.9.4-rhel"}}', nullcontext()),
        (
            "podman",
            '{"Client":{"Version":"5.2.0"},"Server":{"Version":"2.1.2"}}',
            pytest.raises(OCIEngineTooOldError),
        ),
        (
            "podman",
            '{"Client":{"Version":"2.2.0"},"Server":{"Version":"5.1.2"}}',
            pytest.raises(OCIEngineTooOldError),
        ),
        ("podman", '{"Client":{"Version":"3.0~rc1-rhel"}}', nullcontext()),
        ("podman", '{"Client":{"Version":"2.1.0~rc1"}}', pytest.raises(OCIEngineTooOldError)),
    ],
)
def test_engine_version(
    engine_name: str,
    version: str | None,
    context: contextlib.AbstractContextManager[None],
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    def mockcall(*args: object, **kwargs: object) -> str:
        if version is None:
            raise subprocess.CalledProcessError(1, " ".join(str(arg) for arg in args))
        return version

    monkeypatch.setattr(cibuildwheel.oci_container, "call", mockcall)
    engine = OCIContainerEngineConfig.from_config_string(engine_name)
    with context:
        _check_engine_version(engine)



import dataclasses
import io
import json
import os
import platform
import shlex
import shutil
import subprocess
import sys
import textwrap
import typing
import uuid
from collections.abc import Mapping, Sequence
from enum import Enum
from pathlib import Path, PurePath, PurePosixPath
from types import TracebackType
from typing import IO, Literal, Self, assert_never

from cibuildwheel.ci import CIProvider, detect_ci_provider
from cibuildwheel.errors import OCIEngineTooOldError
from cibuildwheel.logger import log
from cibuildwheel.typing import PathOrStr
from cibuildwheel.util.cmd import call
from cibuildwheel.util.helpers import FlexibleVersion, parse_key_value_string, strtobool

ContainerEngineName = Literal["docker", "podman"]


# Order of the enum matters for tests. 386 shall appear before amd64.
class OCIPlatform(Enum):
    i386 = "linux/386"
    AMD64 = "linux/amd64"
    ARMV7 = "linux/arm/v7"
    ARM64 = "linux/arm64"
    PPC64LE = "linux/ppc64le"
    RISCV64 = "linux/riscv64"
    S390X = "linux/s390x"

    @classmethod
    def native(cls) -> Self:
        """Return the current OCI platform, or raise ValueError if unknown."""
        arch = platform.machine().lower()
        mapping = {
            "i386": cls.i386,
            "i686": cls.i386,
            "x86_64": cls.AMD64,
            "amd64": cls.AMD64,
            "armv7l": cls.ARMV7,
            "aarch64": cls.ARM64,
            "arm64": cls.ARM64,
            "ppc64le": cls.PPC64LE,
            "riscv64": cls.RISCV64,
            "s390x": cls.S390X,
        }
        try:
            return mapping[arch]
        except KeyError as ex:
            msg = f"Unsupported platform architecture: {arch}"
            raise OSError(msg) from ex


@dataclasses.dataclass(frozen=True)
class OCIContainerEngineConfig:
    name: ContainerEngineName
    _: dataclasses.KW_ONLY
    create_args: tuple[str, ...] = dataclasses.field(default_factory=tuple)
    disable_host_mount: bool = False

    @classmethod
    def from_config_string(cls, config_string: str) -> Self:
        config_dict = parse_key_value_string(
            config_string,
            ["name"],
            ["create_args", "create-args", "disable_host_mount", "disable-host-mount"],
        )
        name = " ".join(config_dict["name"])
        if name not in {"docker", "podman"}:
            msg = f"unknown container engine {name}"
            raise ValueError(msg)

        name = typing.cast("ContainerEngineName", name)
        # some flexibility in the option names to cope with TOML conventions
        create_args = config_dict.get("create_args") or config_dict.get("create-args") or []
        disable_host_mount_options = (
            config_dict.get("disable_host_mount") or config_dict.get("disable-host-mount") or []
        )
        disable_host_mount = (
            strtobool(disable_host_mount_options[-1]) if disable_host_mount_options else False
        )
        if "--platform" in create_args or any(arg.startswith("--platform=") for arg in create_args):
            msg = "Using '--platform' in 'container-engine::create_args' is deprecated. It will be ignored."
            log.warning(msg)
            if "--platform" in create_args:
                index = create_args.index("--platform")
                create_args.pop(index)
                create_args.pop(index)
            else:
                create_args = [arg for arg in create_args if not arg.startswith("--platform=")]

        return cls(name=name, create_args=tuple(create_args), disable_host_mount=disable_host_mount)

    def options_summary(self) -> str | dict[str, str]:
        if not self.create_args:
            return self.name
        else:
            return {
                "name": self.name,
                "create_args": repr(self.create_args),
                "disable_host_mount": str(self.disable_host_mount),
            }


DEFAULT_ENGINE = OCIContainerEngineConfig("docker")


def _check_engine_version(engine: OCIContainerEngineConfig) -> None:
    try:
        version_string = call(engine.name, "version", "-f", "{{json .}}", capture_stdout=True)
        # We are using lowercase keys for all dicts so we are not affected by casing of keys
        version_info = json.loads(
            version_string.strip(), object_pairs_hook=lambda inp: {k.lower(): v for k, v in inp}
        )
        match engine.name:
            case "docker":
                client_api_version = FlexibleVersion(version_info["client"]["apiversion"])
                server_api_version = FlexibleVersion(version_info["server"]["apiversion"])
                # --platform support was introduced in 1.32 as experimental, 1.41 removed the experimental flag
                version = min(client_api_version, server_api_version)
                minimum_version = FlexibleVersion("1.41")
                minimum_version_str = "20.10.0"  # docker version
                error_msg = textwrap.dedent(
                    f"""
                    Build failed because {engine.name} is too old.

                    cibuildwheel requires {engine.name}>={minimum_version_str} running API version {minimum_version}.
                    The API version found by cibuildwheel is {version}.
                    """
                )
            case "podman":
                # podman uses the same version string for "Version" & "ApiVersion"
                client_version = FlexibleVersion(version_info["client"]["version"])
                if "server" in version_info:
                    server_version = FlexibleVersion(version_info["server"]["version"])
                else:
                    server_version = client_version
                # --platform support was introduced in v3
                version = min(client_version, server_version)
                minimum_version = FlexibleVersion("3")
                error_msg = textwrap.dedent(
                    f"""
                    Build failed because {engine.name} is too old.

                    cibuildwheel requires {engine.name}>={minimum_version}.
                    The version found by cibuildwheel is {version}.
                    """
                )
            case _:
                assert_never(engine.name)

        if version < minimum_version:
            raise OCIEngineTooOldError(error_msg) from None

    except (subprocess.CalledProcessError, KeyError, ValueError) as e:
        msg = f"Build failed because {engine.name} is too old or is not working properly."
        raise OCIEngineTooOldError(msg) from e


class OCIContainer:
    """
    An object that represents a running OCI (e.g. Docker) container.

    Intended for use as a context manager e.g.
    `with OCIContainer(image = 'ubuntu') as docker:`

    A bash shell is running in the remote container. When `call()` is invoked,
    the command is relayed to the remote shell, and the results are streamed
    back to cibuildwheel.

    Example:
        >>> # xdoctest: +REQUIRES(LINUX)
        >>> from cibuildwheel.oci_container import *  # NOQA
        >>> from cibuildwheel.options import _get_pinned_container_images
        >>> import pytest
        >>> try:
        ...     oci_platform = OCIPlatform.native()
        ... except OSError as ex:
        ...     pytest.skip(str(ex))
        >>> if oci_platform != OCIPlatform.AMD64:
        ...     pytest.skip('only runs on amd64')
        >>> image = _get_pinned_container_images()['x86_64']['manylinux2014']
        >>> # Test the default container
        >>> with OCIContainer(image=image, oci_platform=oci_platform) as self:
        ...     self.call(["echo", "hello world"])
        ...     self.call(["cat", "/proc/1/cgroup"])
        ...     print(self.get_environment())
        ...     print(self.debug_info())
    """

    UTILITY_PYTHON = "/opt/python/cp39-cp39/bin/python"

    process: subprocess.Popen[bytes]
    bash_stdin: IO[bytes]
    bash_stdout: IO[bytes]

    def __init__(
        self,
        *,
        image: str,
        oci_platform: OCIPlatform,
        cwd: PathOrStr | None = None,
        engine: OCIContainerEngineConfig = DEFAULT_ENGINE,
    ):
        if not image:
            msg = "Must have a non-empty image to run."
            raise ValueError(msg)

        self.image = image
        self.oci_platform = oci_platform
        self.cwd = cwd
        self.name: str | None = None
        self.engine = engine
        self.host_tar_format = ""
        if sys.platform.startswith("darwin"):
            self.host_tar_format = "--format gnutar"

    def _get_platform_args(self, *, oci_platform: OCIPlatform | None = None) -> tuple[str, str]:
        if oci_platform is None:
            oci_platform = self.oci_platform

        # we need '--pull=always' otherwise some images with the wrong platform get reused (e.g. 386 image for amd64)
        # c.f. https://github.com/moby/moby/issues/48197#issuecomment-2282802313
        pull = "always"
        try:
            image_platform = call(
                self.engine.name,
                "image",
                "inspect",
                self.image,
                "--format",
                (
                    "{{.Os}}/{{.Architecture}}/{{.Variant}}"
                    if len(oci_platform.value.split("/")) == 3
                    else "{{.Os}}/{{.Architecture}}"
                ),
                capture_stdout=True,
            ).strip()
            if image_platform == oci_platform.value:
                # in case the correct image is already present, don't pull
                # this allows to run local only images
                pull = "never"
        except subprocess.CalledProcessError:
            pass
        return f"--platform={oci_platform.value}", f"--pull={pull}"

    def __enter__(self) -> Self:
        self.name = f"cibuildwheel-{uuid.uuid4()}"

        _check_engine_version(self.engine)

        # work-around for Travis-CI PPC64le Docker runs since 2021:
        # this avoids network splits
        # https://github.com/pypa/cibuildwheel/issues/904
        # https://github.com/conda-forge/conda-smithy/pull/1520
        network_args = []
        if detect_ci_provider() == CIProvider.travis_ci and platform.machine() == "ppc64le":
            network_args = ["--network=host"]

        platform_args = self._get_platform_args()

        simulate_32_bit = False
        if self.oci_platform in {OCIPlatform.i386, OCIPlatform.ARMV7}:
            # If the architecture running the image is already the right one
            # or the image entrypoint takes care of enforcing this, then we don't need to
            # simulate this
            run_cmd = [self.engine.name, "run", "--rm"]
            ctr_cmd = ["uname", "-m"]
            try:
                container_machine = call(
                    *run_cmd, *platform_args, self.image, *ctr_cmd, capture_stdout=True
                ).strip()
            except subprocess.CalledProcessError:
                if self.oci_platform == OCIPlatform.i386:
                    # The image might have been built with amd64 architecture
                    # Let's try that
                    platform_args = self._get_platform_args(oci_platform=OCIPlatform.AMD64)
                    container_machine = call(
                        *run_cmd, *platform_args, self.image, *ctr_cmd, capture_stdout=True
                    ).strip()
                else:
                    raise
            simulate_32_bit = container_machine not in {"i686", "armv7l", "armv8l"}

        shell_args = ["linux32", "/bin/bash"] if simulate_32_bit else ["/bin/bash"]

        subprocess.run(
            [
                self.engine.name,
                "create",
                "--env=CIBUILDWHEEL",
                "--env=SOURCE_DATE_EPOCH",
                f"--name={self.name}",
                "--interactive",
                *(["--volume=/:/host"] if not self.engine.disable_host_mount else []),
                *network_args,
                *platform_args,
                *self.engine.create_args,
                self.image,
                *shell_args,
            ],
            check=True,
        )

        self.process = subprocess.Popen(
            [
                self.engine.name,
                "start",
                "--attach",
                "--interactive",
                self.name,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

        assert self.process.stdin
        assert self.process.stdout
        self.bash_stdin = self.process.stdin
        self.bash_stdout = self.process.stdout

        # run a noop command to block until the container is responding
        self.call(["/bin/true"], cwd="/")

        if self.cwd:
            # Although `docker create -w` does create the working dir if it
            # does not exist, podman does not. There does not seem to be a way
            # to setup a workdir for a container running in podman.
            self.call(["mkdir", "-p", os.fspath(self.cwd)], cwd="/")

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.bash_stdin.write(b"exit 0\n")
        self.bash_stdin.flush()
        self.process.wait(timeout=30)
        self.bash_stdin.close()
        self.bash_stdout.close()

        if self.engine.name == "podman":
            # This works around what seems to be a race condition in the podman
            # backend. The full reason is not understood. See PR #966 for a
            # discussion on possible causes and attempts to remove this line.
            # For now, this seems to work "well enough".
            self.process.wait()

        assert isinstance(self.name, str)

        keep_container = strtobool(os.environ.get("CIBW_DEBUG_KEEP_CONTAINER", ""))
        if not keep_container:
            subprocess.run(
                [self.engine.name, "rm", "--force", "-v", self.name],
                stdout=subprocess.DEVNULL,
                check=False,
            )
            self.name = None

    def copy_into(self, from_path: Path, to_path: PurePath) -> None:
        if from_path.is_dir():
            self.call(["mkdir", "-p", to_path])
            subprocess.run(
                f"tar -c {self.host_tar_format} -f - . | {self.engine.name} exec -i {self.name} tar --no-same-owner -xC {shell_quote(to_path)} -f -",
                shell=True,
                check=True,
                cwd=from_path,
            )
        else:
            exec_process: subprocess.Popen[bytes]
            self.call(["mkdir", "-p", to_path.parent])
            with subprocess.Popen(
                [
                    self.engine.name,
                    "exec",
                    "-i",
                    str(self.name),
                    "sh",
                    "-c",
                    f"cat > {shell_quote(to_path)}",
                ],
                stdin=subprocess.PIPE,
            ) as exec_process:
                assert exec_process.stdin
                with open(from_path, "rb") as from_file:
                    shutil.copyfileobj(from_file, exec_process.stdin)

                exec_process.stdin.close()
                exec_process.wait()

                if exec_process.returncode:
                    raise subprocess.CalledProcessError(
                        exec_process.returncode, exec_process.args, None, None
                    )

    def copy_out(self, from_path: PurePath, to_path: Path) -> None:
        # note: we assume from_path is a dir
        to_path.mkdir(parents=True, exist_ok=True)
        call(self.engine.name, "cp", f"{self.name}:{from_path}/.", to_path)

    def glob(self, path: PurePosixPath, pattern: str) -> list[PurePosixPath]:
        glob_pattern = path.joinpath(pattern)

        path_strings = json.loads(
            self.call(
                [
                    self.UTILITY_PYTHON,
                    "-c",
                    f"import sys, json, glob; json.dump(glob.glob({str(glob_pattern)!r}), sys.stdout)",
                ],
                capture_output=True,
            )
        )

        return [PurePosixPath(p) for p in path_strings]

    def call(
        self,
        args: Sequence[PathOrStr],
        env: Mapping[str, str] | None = None,
        capture_output: bool = False,
        cwd: PathOrStr | None = None,
    ) -> str:
        if cwd is None:
            # Podman does not start the a container in a specific working dir
            # so we always need to specify it when making calls.
            cwd = self.cwd

        chdir = f"cd {cwd}" if cwd else ""
        env_assignments = (
            " ".join(f"{shlex.quote(k)}={shlex.quote(v)}" for k, v in env.items())
            if env is not None
            else ""
        )
        command = " ".join(shlex.quote(str(a)) for a in args)
        end_of_message = str(uuid.uuid4())

        # log the command we're executing
        print(f"    + {command}")

        # Write a command to the remote shell. First we change the
        # cwd, if that's required. Then, we use the `env` utility to run
        # `command` inside the specified environment. We use `env` because it
        # can cope with spaces and strange characters in the name or value.
        # Finally, the remote shell is told to write a footer - this will show
        # up in the output so we know when to stop reading, and will include
        # the return code of `command`.
        self.bash_stdin.write(
            bytes(
                f"""(
            {chdir}
            env {env_assignments} {command}
            printf "%04d%s\n" $? {end_of_message}
        )
        """,
                encoding="utf8",
                errors="surrogateescape",
            )
        )
        self.bash_stdin.flush()

        if capture_output:
            output_io: IO[bytes] = io.BytesIO()
        else:
            output_io = sys.stdout.buffer

        while True:
            line = self.bash_stdout.readline()

            if line.endswith(bytes(end_of_message, encoding="utf8") + b"\n"):
                # fmt: off
                footer_offset = (
                    len(line)
                    - 1  # newline character
                    - len(end_of_message)  # delimiter
                    - 4  # 4 return code decimals
                )
                # fmt: on
                return_code_str = line[footer_offset : footer_offset + 4]
                return_code = int(return_code_str)
                # add the last line to output, without the footer
                output_io.write(line[0:footer_offset])
                output_io.flush()
                break
            else:
                output_io.write(line)
                output_io.flush()

        if isinstance(output_io, io.BytesIO):
            output = str(output_io.getvalue(), encoding="utf8", errors="surrogateescape")
        else:
            output = ""

        if return_code != 0:
            raise subprocess.CalledProcessError(return_code, args, output)

        return output

    def get_environment(self) -> dict[str, str]:
        env = json.loads(
            self.call(
                [
                    self.UTILITY_PYTHON,
                    "-c",
                    "import sys, json, os; json.dump(os.environ.copy(), sys.stdout)",
                ],
                capture_output=True,
            )
        )
        return typing.cast("dict[str, str]", env)

    def environment_executor(self, command: Sequence[str], environment: dict[str, str]) -> str:
        # used as an EnvironmentExecutor to evaluate commands and capture output
        return self.call(command, env=environment, capture_output=True)

    def debug_info(self) -> str:
        if self.engine.name == "podman":
            command = f"{self.engine.name} info --debug"
        else:
            command = f"{self.engine.name} info"
        completed = subprocess.run(
            command,
            shell=True,
            check=True,
            cwd=self.cwd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        output = str(completed.stdout, encoding="utf8", errors="surrogateescape")
        return output


def shell_quote(path: PurePath) -> str:
    return shlex.quote(os.fspath(path))



import dataclasses
import itertools
import os
import re
import shlex
import textwrap
from collections import defaultdict
from collections.abc import Sequence

from cibuildwheel.typing import PathOrStr


def format_safe(template: str, **kwargs: str | os.PathLike[str]) -> str:
    """
    Works similarly to `template.format(**kwargs)`, except that unmatched
    fields in `template` are passed through untouched.

    >>> format_safe('{a} {b}', a='123')
    '123 {b}'
    >>> format_safe('{a} {b[4]:3f}', a='123')
    '123 {b[4]:3f}'

    To avoid variable expansion, precede with a single hash e.g.
    >>> format_safe('#{a} {b}', a='123')
    '{a} {b}'
    """

    result = template

    for key, value in kwargs.items():
        find_pattern = re.compile(
            rf"""
                (?<!\#)  # don't match if preceded by a hash
                {{  # literal open curly bracket
                {re.escape(key)}  # the field name
                }}  # literal close curly bracket
            """,
            re.VERBOSE,
        )

        result = re.sub(
            pattern=find_pattern,
            repl=str(value).replace("\\", r"\\"),
            string=result,
        )

        # transform escaped sequences into their literal equivalents
        result = result.replace(f"#{{{key}}}", f"{{{key}}}")

    return result


def prepare_command(command: str, **kwargs: PathOrStr) -> str:
    """
    Preprocesses a command by expanding variables like {project}.

    For example, used in the test_command option to specify the path to the
    project's root. Unmatched syntax will mostly be allowed through.
    """
    return format_safe(command, **kwargs)


def strtobool(val: str) -> bool:
    return val.lower() in {"y", "yes", "t", "true", "on", "1"}


def unwrap(text: str) -> str:
    """
    Unwraps multi-line text to a single line
    """
    # remove initial line indent
    text = textwrap.dedent(text)
    # remove leading/trailing whitespace
    text = text.strip()
    # remove consecutive whitespace
    return re.sub(r"\s+", " ", text)


def unwrap_preserving_paragraphs(text: str) -> str:
    """
    Unwraps multi-line text to a single line, but preserves paragraphs
    """
    # remove initial line indent
    text = textwrap.dedent(text)
    # remove leading/trailing whitespace
    text = text.strip()

    paragraphs = text.split("\n\n")
    # remove consecutive whitespace
    paragraphs = [re.sub(r"\s+", " ", paragraph) for paragraph in paragraphs]
    return "\n\n".join(paragraphs)


def parse_key_value_string(
    key_value_string: str,
    positional_arg_names: Sequence[str] | None = None,
    kw_arg_names: Sequence[str] | None = None,
) -> dict[str, list[str]]:
    """
    Parses a string like "docker; create_args: --some-option=value another-option"
    """
    if positional_arg_names is None:
        positional_arg_names = []
    if kw_arg_names is None:
        kw_arg_names = []

    all_field_names = [*positional_arg_names, *kw_arg_names]

    shlexer = shlex.shlex(key_value_string, posix=True, punctuation_chars=";")
    shlexer.commenters = ""
    shlexer.whitespace_split = True
    parts = list(shlexer)
    # parts now looks like
    # ['docker', ';', 'create_args:', '--some-option=value', 'another-option']

    # split by semicolon
    fields = [list(group) for k, group in itertools.groupby(parts, lambda x: x == ";") if not k]

    result: defaultdict[str, list[str]] = defaultdict(list)
    for field_i, field in enumerate(fields):
        # check to see if the option name is specified
        field_name, sep, first_value = field[0].partition(":")
        if sep:
            if field_name not in all_field_names:
                msg = f"Failed to parse {key_value_string!r}. Unknown field name {field_name!r}"
                raise ValueError(msg)

            values = ([first_value] if first_value else []) + field[1:]
        else:
            try:
                field_name = positional_arg_names[field_i]
            except IndexError:
                msg = f"Failed to parse {key_value_string!r}. Too many positional arguments - expected a maximum of {len(positional_arg_names)}"
                raise ValueError(msg) from None

            values = field

        result[field_name] += values

    return dict(result)


@dataclasses.dataclass(order=True)
class FlexibleVersion:
    version_parts: tuple[int, ...] = dataclasses.field(init=False, repr=False)
    suffix: str = dataclasses.field(init=False, repr=False)
    version_str: str = dataclasses.field(compare=False)

    def __post_init__(self) -> None:
        # Split into numeric parts and the optional suffix
        match = re.match(r"^[v]?(\d+(\.\d+)*)(.*)$", self.version_str)
        if not match:
            msg = f"Invalid version string: {self.version_str}"
            raise ValueError(msg)

        version_part, _, suffix = match.groups()

        # Convert numeric version part into a tuple of integers
        self.version_parts = tuple(map(int, version_part.split(".")))
        self.suffix = suffix.strip() if suffix else ""

        # Normalize by removing trailing zeros
        self.version_parts = self._remove_trailing_zeros(self.version_parts)

    @staticmethod
    def _remove_trailing_zeros(parts: tuple[int, ...]) -> tuple[int, ...]:
        # Remove trailing zeros for accurate comparisons
        # without this, "3.0" would be considered greater than "3"
        while parts and parts[-1] == 0:
            parts = parts[:-1]
        return parts

    def __str__(self) -> str:
        return self.version_str



import csv
import hashlib
import os
import platform
import re
import shlex
import shutil
import subprocess
import sysconfig
from collections.abc import Iterable, Iterator, MutableMapping
from dataclasses import dataclass
from os.path import relpath
from pathlib import Path
from pprint import pprint
from runpy import run_path
from textwrap import dedent
from typing import Any

from build import ProjectBuilder
from build.env import IsolatedEnv
from elftools.common.exceptions import ELFError
from elftools.elf.elffile import ELFFile
from filelock import FileLock

from cibuildwheel import errors, platforms  # pylint: disable=cyclic-import
from cibuildwheel.architecture import Architecture, arch_synonym
from cibuildwheel.frontend import get_build_frontend_extra_flags, parse_config_settings
from cibuildwheel.logger import log
from cibuildwheel.options import BuildOptions, Options
from cibuildwheel.selector import BuildSelector
from cibuildwheel.util import resources
from cibuildwheel.util.cmd import call, shell
from cibuildwheel.util.file import CIBW_CACHE_PATH, copy_test_sources, download, move_file
from cibuildwheel.util.helpers import prepare_command
from cibuildwheel.util.packaging import find_compatible_wheel
from cibuildwheel.util.python_build_standalone import create_python_build_standalone_environment
from cibuildwheel.venv import constraint_flags, find_uv, virtualenv

ANDROID_TRIPLET = {
    "arm64_v8a": "aarch64-linux-android",
    "x86_64": "x86_64-linux-android",
}


def parse_identifier(identifier: str) -> tuple[str, str]:
    match = re.fullmatch(r"cp(\d)(\d+)-android_(.+)", identifier)
    if not match:
        msg = f"invalid Android identifier: '{identifier}'"
        raise ValueError(msg)
    major, minor, arch = match.groups()
    return (f"{major}.{minor}", arch)


def android_triplet(identifier: str) -> str:
    return ANDROID_TRIPLET[parse_identifier(identifier)[1]]


@dataclass(frozen=True)
class PythonConfiguration:
    version: str
    identifier: str
    url: str

    @property
    def arch(self) -> str:
        return parse_identifier(self.identifier)[1]


def all_python_configurations() -> list[PythonConfiguration]:
    return [PythonConfiguration(**item) for item in resources.read_python_configs("android")]


def get_python_configurations(
    build_selector: BuildSelector, architectures: set[Architecture]
) -> list[PythonConfiguration]:
    return [
        c
        for c in all_python_configurations()
        if c.arch in architectures and build_selector(c.identifier)
    ]


def shell_prepared(command: str, *, build_options: BuildOptions, env: dict[str, str]) -> None:
    shell(
        prepare_command(command, project=".", package=build_options.package_dir),
        env=env,
    )


def before_all(options: Options, python_configurations: list[PythonConfiguration]) -> None:
    before_all_options = options.build_options(python_configurations[0].identifier)
    if before_all_options.before_all:
        log.step("Running before_all...")
        shell_prepared(
            before_all_options.before_all,
            build_options=before_all_options,
            env=before_all_options.environment.as_dictionary(os.environ),
        )


@dataclass(frozen=True)
class BuildState:
    config: PythonConfiguration
    options: BuildOptions
    build_path: Path
    python_dir: Path
    build_env: dict[str, str]
    android_env: dict[str, str]


def build(options: Options, tmp_path: Path) -> None:
    if "ANDROID_HOME" not in os.environ:
        msg = (
            "ANDROID_HOME environment variable is not set. For instructions, see "
            "https://cibuildwheel.pypa.io/en/stable/platforms/#android"
        )
        raise errors.FatalError(msg)

    configs = get_python_configurations(
        options.globals.build_selector, options.globals.architectures
    )
    if not configs:
        return

    try:
        before_all(options, configs)

        built_wheels: list[Path] = []
        for config in configs:
            log.build_start(config.identifier)
            build_options = options.build_options(config.identifier)
            build_path = tmp_path / config.identifier
            build_path.mkdir()
            python_dir = setup_target_python(config, build_path)
            build_env, android_env = setup_env(config, build_options, build_path, python_dir)
            state = BuildState(
                config, build_options, build_path, python_dir, build_env, android_env
            )

            compatible_wheel = find_compatible_wheel(built_wheels, config.identifier)
            if compatible_wheel:
                print(
                    f"\nFound previously built wheel {compatible_wheel.name} that is "
                    f"compatible with {config.identifier}. Skipping build step..."
                )
                repaired_wheel = compatible_wheel
            else:
                before_build(state)
                built_wheel = build_wheel(state)
                repaired_wheel = repair_wheel(state, built_wheel)

            test_wheel(state, repaired_wheel, build_frontend=build_options.build_frontend.name)

            output_wheel: Path | None = None
            if compatible_wheel is None:
                output_wheel = move_file(
                    repaired_wheel, build_options.output_dir / repaired_wheel.name
                )
                built_wheels.append(output_wheel)

            shutil.rmtree(build_path)
            log.build_end(output_wheel)

    except subprocess.CalledProcessError as error:
        msg = f"Command {error.cmd} failed with code {error.returncode}. {error.stdout or ''}"
        raise errors.FatalError(msg) from error


def setup_target_python(config: PythonConfiguration, build_path: Path) -> Path:
    log.step("Installing target Python...")
    python_tgz = CIBW_CACHE_PATH / config.url.rpartition("/")[-1]
    with FileLock(f"{python_tgz}.lock"):
        if not python_tgz.exists():
            download(config.url, python_tgz)

    python_dir = build_path / "python"
    python_dir.mkdir()
    shutil.unpack_archive(python_tgz, python_dir)
    return python_dir


def setup_env(
    config: PythonConfiguration, build_options: BuildOptions, build_path: Path, python_dir: Path
) -> tuple[dict[str, str], dict[str, str]]:
    """
    Returns two environment dicts, both pointing at the same virtual environment:

    * build_env, which uses the environment normally.
    * android_env, which uses the environment while simulating running on Android.
    """
    log.step("Setting up build environment...")
    build_frontend = build_options.build_frontend.name
    use_uv = build_frontend in {"build[uv]", "uv"}
    uv_path = find_uv()
    if use_uv and uv_path is None:
        msg = "uv not found"
        raise AssertionError(msg)
    pip = ["pip"] if not use_uv else [str(uv_path), "pip"]

    # Create virtual environment
    python_exe = create_python_build_standalone_environment(
        config.version, build_path, CIBW_CACHE_PATH
    )
    venv_dir = build_path / "venv"
    dependency_constraint = build_options.dependency_constraints.get_for_python_version(
        version=config.version, tmp_dir=build_path
    )
    build_env = virtualenv(
        config.version, python_exe, venv_dir, dependency_constraint, use_uv=use_uv
    )
    create_cmake_toolchain(config, build_path, python_dir, build_env)

    # Apply custom environment variables, and check environment is still valid
    build_env = build_options.environment.as_dictionary(build_env)
    build_env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"
    for command in ["python"] if use_uv else ["python", "pip"]:
        command_path = call("which", command, env=build_env, capture_stdout=True).strip()
        if command_path != f"{venv_dir}/bin/{command}":
            msg = (
                f"{command} available on PATH doesn't match our installed instance. If you "
                f"have modified PATH, ensure that you don't overwrite cibuildwheel's entry "
                f"or insert {command} above it."
            )
            raise errors.FatalError(msg)
        call(command, "--version", env=build_env)

    # Construct an altered environment which simulates running on Android.
    android_env = setup_android_env(config, python_dir, venv_dir, build_env)

    # Install build tools
    if build_frontend not in {"build", "build[uv]", "uv"}:
        msg = "Android requires the build frontend to be 'build' or 'uv'"
        raise errors.FatalError(msg)
    call(*pip, "install", "build", *constraint_flags(dependency_constraint), env=build_env)

    # Build-time requirements must be queried within android_env, because
    # `get_requires_for_build` can run arbitrary code in setup.py scripts, which may be
    # affected by the target platform. However, the requirements must be installed
    # within build_env, because they're going to run on the build machine.
    #
    # The `build` CLI doesn't support this combination, so we use its API to query the
    # requirements, and then install them ourselves with pip. We'll later run `build` in
    # the same environment, passing the `--no-isolation` option.
    class AndroidEnv(IsolatedEnv):
        @property
        def python_executable(self) -> str:
            return f"{venv_dir}/bin/python"

        def make_extra_environ(self) -> dict[str, str]:
            return android_env

    pb = ProjectBuilder.from_isolated_env(AndroidEnv(), build_options.package_dir)
    if pb.build_system_requires:
        call(*pip, "install", *pb.build_system_requires, env=build_env)

    requires_for_build = pb.get_requires_for_build(
        "wheel", parse_config_settings(build_options.config_settings)
    )
    if requires_for_build:
        call(*pip, "install", *requires_for_build, env=build_env)

    return build_env, android_env


def create_cmake_toolchain(
    config: PythonConfiguration, build_path: Path, python_dir: Path, build_env: dict[str, str]
) -> None:
    toolchain_path = build_path / "toolchain.cmake"
    build_env["CMAKE_TOOLCHAIN_FILE"] = str(toolchain_path)
    with open(toolchain_path, "w", encoding="UTF-8") as toolchain_file:
        print(
            dedent(
                f"""\
                # To support as many build systems as possible, we use environment
                # variables as the single source of truth for compiler flags and paths,
                # so they don't need to be specified here.

                set(CMAKE_SYSTEM_NAME Android)
                set(CMAKE_SYSTEM_PROCESSOR {android_triplet(config.identifier).split("-")[0]})

                # Inhibit all of CMake's own NDK handling code.
                set(CMAKE_SYSTEM_VERSION 1)

                # Tell CMake where to look for headers and libraries.
                set(CMAKE_FIND_ROOT_PATH "{python_dir}/prefix")
                set(CMAKE_FIND_ROOT_PATH_MODE_PROGRAM NEVER)
                set(CMAKE_FIND_ROOT_PATH_MODE_LIBRARY ONLY)
                set(CMAKE_FIND_ROOT_PATH_MODE_INCLUDE ONLY)
                set(CMAKE_FIND_ROOT_PATH_MODE_PACKAGE BOTH)

                # Allow CMake to run Python in the simulated Android environment when
                # policy CMP0190 is active.
                set(CMAKE_CROSSCOMPILING_EMULATOR /bin/sh -c [["$0" "$@"]])
                """
            ),
            file=toolchain_file,
        )


def localize_sysconfigdata(
    python_dir: Path, build_env: dict[str, str], sysconfigdata_path: Path
) -> dict[str, Any]:
    sysconfigdata: dict[str, Any] = run_path(str(sysconfigdata_path))["build_time_vars"]
    with sysconfigdata_path.open("w", encoding="UTF-8") as f:
        f.write("# Generated by cibuildwheel\n")
        f.write("build_time_vars = ")
        sysconfigdata = localized_vars(build_env, sysconfigdata, python_dir / "prefix")
        pprint(sysconfigdata, stream=f, compact=True)
        return sysconfigdata


def localized_vars(
    build_env: dict[str, str], orig_vars: dict[str, Any], prefix: Path
) -> dict[str, Any]:
    orig_prefix = orig_vars["prefix"]
    localized_vars_ = {}
    for key, value in orig_vars.items():
        # The host's sysconfigdata will include references to build-time paths.
        # Update these to refer to the current prefix.
        final = value
        if isinstance(final, str):
            final = final.replace(orig_prefix, str(prefix))

        if key == "ANDROID_API_LEVEL":
            if api_level := build_env.get(key):
                final = int(api_level)

        # Build systems vary in whether FLAGS variables are read from sysconfig, and if so,
        # whether they're replaced by environment variables or combined with them. Even
        # setuptools has changed its behavior here
        # (https://github.com/pypa/setuptools/issues/4836).
        #
        # Ensure consistency by clearing the sysconfig variables and letting the environment
        # variables take effect alone. This will also work for any non-Python build systems
        # which the build script may call.
        elif key in ["CFLAGS", "CXXFLAGS", "LDFLAGS"]:
            final = ""

        # These variables contain an embedded copy of LDFLAGS.
        elif key in ["LDSHARED", "LDCXXSHARED"]:
            final = final.removesuffix(" " + orig_vars["LDFLAGS"])

        localized_vars_[key] = final

    return localized_vars_


def setup_android_env(
    config: PythonConfiguration, python_dir: Path, venv_dir: Path, build_env: dict[str, str]
) -> dict[str, str]:
    site_packages = next(venv_dir.glob("lib/python*/site-packages"))
    for suffix in ["pth", "py"]:
        shutil.copy(resources.PATH / f"_cross_venv.{suffix}", site_packages)

    sysconfigdata_path = Path(
        shutil.copy(
            next(python_dir.glob("prefix/lib/python*/_sysconfigdata_*.py")),
            site_packages,
        )
    )
    sysconfigdata = localize_sysconfigdata(python_dir, build_env, sysconfigdata_path)

    # Activate the code in _cross_venv.py.
    android_env = build_env.copy()
    android_env["CIBW_HOST_TRIPLET"] = android_triplet(config.identifier)

    # Get the environment variables needed to build for Android (CC, CFLAGS, etc). These are
    # generated by https://github.com/python/cpython/blob/main/Android/android-env.sh.
    env_output = call(python_dir / "android.py", "env", env=build_env, capture_stdout=True)

    # shlex.split should produce a sequence alternating between:
    #   * the word "export"
    #   * a key=value string, without quotes
    for i, token in enumerate(shlex.split(env_output)):
        if i % 2 == 0:
            assert token == "export", token
        else:
            key, sep, value = token.partition("=")
            assert sep == "=", token
            android_env[key] = value

    # localized_vars cleared the CFLAGS and CXXFLAGS in the sysconfigdata, but most
    # packages take their optimization flags from these variables. Pass these flags via
    # environment variables instead.
    #
    # We don't enable debug information, because it significantly increases binary size,
    # and most Android app developers don't have the NDK installed, so they would have no
    # way to strip it.
    opt = " ".join(word for word in sysconfigdata["OPT"].split() if not word.startswith("-g"))
    for key in ["CFLAGS", "CXXFLAGS"]:
        android_env[key] += " " + opt

    # Cargo target linker needs to be specified after CC is set
    setup_rust(config, python_dir, android_env)

    # Format the environment so it can be pasted into a shell when debugging.
    for key, value in sorted(android_env.items()):
        if os.environ.get(key) != value:
            print(f"export {key}={shlex.quote(value)}")

    return android_env


def setup_rust(
    config: PythonConfiguration,
    python_dir: Path,
    env: MutableMapping[str, str],
) -> None:
    cargo_target = android_triplet(config.identifier)

    # CARGO_BUILD_TARGET is the variable used by Cargo and setuptools_rust
    env["CARGO_BUILD_TARGET"] = cargo_target

    # The linker needs to be specified after CC is set by android-env.sh
    cargo_target_linker_env_name = f"CARGO_TARGET_{cargo_target.upper().replace('-', '_')}_LINKER"
    # CC has already been set by calling android.py (it calls android-env.sh)
    env[cargo_target_linker_env_name] = env["CC"]

    # All Python extension modules must be explicitly linked against libpython3.x.so when building for Android.
    # See: https://peps.python.org/pep-0738/#linkage
    # For projects using PyO3, this requires setting PYO3_CROSS_LIB_DIR to the directory containing libpython3.x.so.
    # See: https://pyo3.rs/v0.27.1/building-and-distribution.html#cross-compiling
    env["PYO3_CROSS_LIB_DIR"] = str(python_dir / "prefix" / "lib")

    venv_bin = Path(env["VIRTUAL_ENV"]) / "bin"
    for tool in ["cargo", "rustup"]:
        shim_path = venv_bin / tool
        shutil.copy(resources.PATH / "_rust_shim.py", shim_path)
        shim_path.chmod(0o755)


def before_build(state: BuildState) -> None:
    if state.options.before_build:
        log.step("Running before_build...")
        shell_prepared(
            state.options.before_build,
            build_options=state.options,
            env=state.build_env,
        )


def build_wheel(state: BuildState) -> Path:
    log.step("Building wheel...")
    built_wheel_dir = state.build_path / "built_wheel"
    match state.options.build_frontend.name:
        case "build" | "build[uv]":
            call(
                "python",
                "-m",
                "build",
                state.options.package_dir,
                "--wheel",
                "--no-isolation",
                "--skip-dependency-check",
                f"--outdir={built_wheel_dir}",
                *get_build_frontend_extra_flags(
                    state.options.build_frontend,
                    state.options.build_verbosity,
                    state.options.config_settings,
                    py38=False,
                ),
                env=state.android_env,
            )
        case "uv":
            uv_path = find_uv()
            assert uv_path is not None
            call(
                uv_path,
                "build",
                state.options.package_dir,
                "--wheel",
                "--no-build-isolation",
                f"--out-dir={built_wheel_dir}",
                *get_build_frontend_extra_flags(
                    state.options.build_frontend,
                    state.options.build_verbosity,
                    state.options.config_settings,
                    py38=False,
                ),
                env=state.android_env,
            )
        case x:
            msg = f"Invalid build backend {x!r}"
            raise AssertionError(msg)

    built_wheels = list(built_wheel_dir.glob("*.whl"))
    if len(built_wheels) != 1:
        msg = f"{built_wheel_dir} contains {len(built_wheels)} wheels; expected 1"
        raise errors.FatalError(msg)
    built_wheel = built_wheels[0]

    if built_wheel.name.endswith("none-any.whl"):
        raise errors.NonPlatformWheelError()
    return built_wheel


def repair_wheel(state: BuildState, built_wheel: Path) -> Path:
    log.step("Repairing wheel...")
    repaired_wheel_dir = state.build_path / "repaired_wheel"
    repaired_wheel_dir.mkdir()

    if state.options.repair_command:
        shell(
            prepare_command(
                state.options.repair_command,
                wheel=built_wheel,
                dest_dir=repaired_wheel_dir,
                package=state.options.package_dir,
                project=".",
            ),
            env=state.build_env,
        )
    else:
        repair_default(state.android_env, built_wheel, repaired_wheel_dir)

    repaired_wheels = list(repaired_wheel_dir.glob("*.whl"))
    if len(repaired_wheels) == 0:
        raise errors.RepairStepProducedNoWheelError()
    if len(repaired_wheels) != 1:
        msg = f"{repaired_wheel_dir} contains {len(repaired_wheels)} wheels; expected 1"
        raise errors.FatalError(msg)
    repaired_wheel = repaired_wheels[0]

    if repaired_wheel.name.endswith("none-any.whl"):
        raise errors.NonPlatformWheelError()
    return repaired_wheel


def repair_default(
    android_env: dict[str, str], built_wheel: Path, repaired_wheel_dir: Path
) -> None:
    """
    Adds libc++ to the wheel if anything links against it. In the future this should be
    moved to auditwheel and generalized to support more libraries.
    """
    if (match := re.search(r"^(.+?)-", built_wheel.name)) is None:
        msg = f"Failed to parse wheel filename: {built_wheel.name}"
        raise errors.FatalError(msg)
    wheel_name = match[1]

    unpacked_dir = repaired_wheel_dir / "unpacked"
    unpacked_dir.mkdir()
    shutil.unpack_archive(built_wheel, unpacked_dir, format="zip")

    # Some build systems are inconsistent about name normalization, so don't assume the
    # dist-info name is identical to the wheel name.
    record_paths = list(unpacked_dir.glob("*.dist-info/RECORD"))
    if len(record_paths) != 1:
        msg = f"{built_wheel.name} contains {len(record_paths)} dist-info/RECORD files; expected 1"
        raise errors.FatalError(msg)

    old_soname = "libc++_shared.so"
    paths_to_patch = []
    for path, elffile in elf_file_filter(
        unpacked_dir / filename
        for filename, *_ in csv.reader(record_paths[0].read_text().splitlines())
    ):
        if (dynamic := elffile.get_section_by_name(".dynamic")) and any(  # type: ignore[no-untyped-call]
            tag.entry.d_tag == "DT_NEEDED" and tag.needed == old_soname
            for tag in dynamic.iter_tags()
        ):
            paths_to_patch.append(path)

    if not paths_to_patch:
        shutil.copyfile(built_wheel, repaired_wheel_dir / built_wheel.name)
    else:
        # Android doesn't support DT_RPATH, but supports DT_RUNPATH since API level 24
        # (https://github.com/aosp-mirror/platform_bionic/blob/master/android-changes-for-ndk-developers.md).
        if int(sysconfig_print('get_config_vars()["ANDROID_API_LEVEL"]', android_env)) < 24:
            msg = f"Adding {old_soname} requires ANDROID_API_LEVEL to be at least 24"
            raise errors.FatalError(msg)

        toolchain = Path(android_env["CC"]).parent.parent
        src_path = toolchain / f"sysroot/usr/lib/{android_env['CIBW_HOST_TRIPLET']}/{old_soname}"

        # Use the same library location as auditwheel would.
        libs_dir = unpacked_dir / (wheel_name + ".libs")
        libs_dir.mkdir()
        new_soname = soname_with_hash(src_path)
        dst_path = libs_dir / new_soname
        shutil.copyfile(src_path, dst_path)
        call(which("patchelf"), "--set-soname", new_soname, dst_path)

        for path in paths_to_patch:
            call(which("patchelf"), "--replace-needed", old_soname, new_soname, path)
            call(
                which("patchelf"),
                "--set-rpath",
                f"${{ORIGIN}}/{relpath(libs_dir, path.parent)}",
                path,
            )
        call(which("wheel"), "pack", unpacked_dir, "-d", repaired_wheel_dir)


# If cibuildwheel was called without activating its environment, its scripts directory
# will not be on the PATH.
def which(cmd: str) -> str:
    scripts_dir = sysconfig.get_path("scripts")
    result = shutil.which(cmd, path=scripts_dir + os.pathsep + os.environ["PATH"])
    if result is None:
        msg = f"Couldn't find {cmd!r} in {scripts_dir} or on the PATH"
        raise errors.FatalError(msg)
    return result


def elf_file_filter(paths: Iterable[Path]) -> Iterator[tuple[Path, ELFFile]]:
    """Filter through an iterator of filenames and load up only ELF files"""
    for path in paths:
        if not path.name.endswith(".py"):
            try:
                with open(path, "rb") as f:
                    candidate = ELFFile(f)  # type: ignore[no-untyped-call]
                    yield path, candidate
            except ELFError:
                pass  # Not an ELF file


def soname_with_hash(src_path: Path) -> str:
    """Return the same library filename as auditwheel would"""
    shorthash = hashlib.sha256(src_path.read_bytes()).hexdigest()[:8]
    src_name = src_path.name
    base, ext = src_name.split(".", 1)
    if not base.endswith(f"-{shorthash}"):
        return f"{base}-{shorthash}.{ext}"
    else:
        return src_name


def test_wheel(state: BuildState, wheel: Path, *, build_frontend: str) -> None:
    test_command = state.options.test_command
    if not (test_command and state.options.test_selector(state.config.identifier)):
        return

    log.step("Testing wheel...")
    use_uv = build_frontend == "build[uv]"
    uv_path = find_uv()
    if use_uv and uv_path is None:
        msg = "uv not found"
        raise AssertionError(msg)
    pip = ["pip"] if not use_uv else [str(uv_path), "pip"]

    native_arch = arch_synonym(platform.machine(), platforms.native_platform(), "android")
    if state.config.arch != native_arch:
        log.warning(
            f"Skipping tests for {state.config.arch}, as the build machine only "
            f"supports {native_arch}"
        )
        return

    if state.options.before_test:
        shell_prepared(
            state.options.before_test,
            build_options=state.options,
            env=state.build_env,
        )

    platform_args = (
        ["--python-platform", android_triplet(state.config.identifier)]
        if use_uv
        else [
            "--platform",
            sysconfig_print("get_platform()", state.android_env).replace("-", "_"),
        ]
    )

    # Install the wheel and test-requires.
    site_packages_dir = state.build_path / "site-packages"
    site_packages_dir.mkdir()
    call(
        *pip,
        "install",
        "--only-binary=:all:",
        *platform_args,
        "--target",
        site_packages_dir,
        f"{wheel}{state.options.test_extras}",
        *state.options.test_requires,
        env=state.android_env,
    )

    # Copy test-sources.
    cwd_dir = state.build_path / "cwd"
    cwd_dir.mkdir()
    if state.options.test_sources:
        copy_test_sources(state.options.test_sources, Path.cwd(), cwd_dir)
    else:
        (cwd_dir / "test_fail.py").write_text(
            resources.TEST_FAIL_CWD_FILE.read_text(),
        )

    # Android doesn't support placeholders in the test command.
    if any(("{" + placeholder + "}") in test_command for placeholder in ("project", "package")):
        msg = (
            f"Test command {test_command!r} with a "
            "'{project}' or '{package}' placeholder is not supported on Android, "
            "because the source directory is not visible on the emulator."
        )
        raise errors.FatalError(msg)

    # Parse test-command.
    test_args = shlex.split(test_command)
    if test_args[0] in {"python", "python3"} and any(arg in test_args for arg in ("-c", "-m")):
        # Forward the args to the CPython testbed script. We require '-c' or '-m'
        # to be in the command, because without those flags, the testbed script
        # will prepend '-m test', which will run Python's own test suite.
        del test_args[0]
    elif test_args[0] == "pytest":
        # We transform some commands into the `python -m` form, but this is deprecated.
        msg = (
            f"Test command {test_command!r} is not supported on Android. "
            "cibuildwheel will try to execute it as if it started with 'python -m'. "
            "If this works, all you need to do is add that to your test command."
        )
        log.warning(msg)
        test_args.insert(0, "-m")
    else:
        msg = (
            f"Test command {test_command!r} is not supported on Android. "
            f"Command must begin with 'python' or 'python3', and contain '-m' or '-c'."
        )
        raise errors.FatalError(msg)

    # By default, run on a testbed managed emulator running the newest supported
    # Android version. However, if the user specifies a --managed or --connected
    # test execution argument, that argument takes precedence.
    test_runtime_args = state.options.test_runtime.args

    if any(arg.startswith(("--managed", "--connected")) for arg in test_runtime_args):
        emulator_args = []
    else:
        emulator_args = ["--managed", "maxVersion"]

    # Run the test app.
    call(
        state.python_dir / "android.py",
        "test",
        "--site-packages",
        site_packages_dir,
        "--cwd",
        cwd_dir,
        *emulator_args,
        *(["-v"] if state.options.build_verbosity > 0 else []),
        *test_runtime_args,
        "--",
        *test_args,
        env=state.build_env,
    )


def sysconfig_print(method_call: str, env: dict[str, str]) -> str:
    return call(
        "python",
        "-c",
        f'import sysconfig; print(sysconfig.{method_call}, end="")',
        env=env,
        capture_stdout=True,
    )



import os
import shlex
import shutil
import subprocess
import sys
import typing
from collections.abc import Iterator, Mapping
from typing import Final, Literal

from cibuildwheel.errors import FatalError
from cibuildwheel.typing import PathOrStr

_IS_WIN: Final[bool] = sys.platform.startswith("win")


@typing.overload
def call(
    *args: PathOrStr,
    env: Mapping[str, str] | None = None,
    cwd: PathOrStr | None = None,
    capture_stdout: Literal[False] = ...,
) -> None: ...


@typing.overload
def call(
    *args: PathOrStr,
    env: Mapping[str, str] | None = None,
    cwd: PathOrStr | None = None,
    capture_stdout: Literal[True],
) -> str: ...


def call(
    *args: PathOrStr,
    env: Mapping[str, str] | None = None,
    cwd: PathOrStr | None = None,
    capture_stdout: bool = False,
) -> str | None:
    """
    Run subprocess.run, but print the commands first. Takes the commands as
    *args. Uses shell=True on Windows due to a bug. Also converts to
    Paths to strings, due to Windows behavior at least on older Pythons.
    https://bugs.python.org/issue8557
    """
    args_ = [str(arg) for arg in args]
    # print the command executing for the logs
    print("+ " + " ".join(shlex.quote(a) for a in args_))
    # workaround platform behaviour differences outlined
    # in https://github.com/python/cpython/issues/52803
    path_env = env if env is not None else os.environ
    path = path_env.get("PATH", None)
    executable = shutil.which(args_[0], path=path)
    if executable is None:
        msg = f"Couldn't find {args_[0]!r} in PATH {path!r}"
        raise FatalError(msg)
    args_[0] = executable
    try:
        result = subprocess.run(
            args_,
            check=True,
            shell=_IS_WIN,
            env=env,
            cwd=cwd,
            capture_output=capture_stdout,
            text=capture_stdout,
        )
    except subprocess.CalledProcessError as e:
        if capture_stdout:
            sys.stderr.write(e.stderr)
        raise
    if not capture_stdout:
        return None
    sys.stderr.write(result.stderr)
    return typing.cast("str", result.stdout)


def shell(
    *commands: str, env: Mapping[str, str] | None = None, cwd: PathOrStr | None = None
) -> None:
    command = " ".join(commands)
    print(f"+ {command}")
    subprocess.run(command, env=env, cwd=cwd, shell=True, check=True)


def split_command(lst: list[str]) -> Iterator[list[str]]:
    """
    Split a shell-style command, as returned by shlex.split, into a sequence
    of commands, separated by '&&'.
    """
    items = list[str]()
    for item in lst:
        if item == "&&":
            yield items
            items = []
        else:
            items.append(item)
    yield items



#!/usr/bin/env python3


import dataclasses
import difflib
import logging
import tomllib
from pathlib import Path
from typing import Final

import click
import packaging.specifiers
import requests
import rich
from packaging.version import InvalidVersion, Version
from rich.logging import RichHandler
from rich.syntax import Syntax

log = logging.getLogger("cibw")

# Looking up the dir instead of using utils.resources_dir
# since we want to write to it.
DIR: Final[Path] = Path(__file__).parent.parent.resolve()
RESOURCES_DIR: Final[Path] = DIR / "cibuildwheel/resources"

NODEJS_DIST: Final[str] = "https://nodejs.org/dist/"
NODEJS_INDEX: Final[str] = f"{NODEJS_DIST}index.json"


@dataclasses.dataclass(frozen=True, order=True)
class VersionTuple:
    version: Version
    version_string: str


def parse_nodejs_index() -> list[VersionTuple]:
    versions: list[VersionTuple] = []
    response = requests.get(NODEJS_INDEX)
    response.raise_for_status()
    versions_info = response.json()
    for version_info in versions_info:
        version_string = version_info.get("version", "???")
        if not version_info.get("lts", False):
            log.debug("Ignoring non LTS release %r", version_string)
            continue
        if "linux-x64" not in version_info.get("files", []):
            log.warning(
                "Ignoring release %r which does not include a linux-x64 binary", version_string
            )
            continue
        try:
            version = Version(version_string)
            if version.is_devrelease:
                log.info("Ignoring development release %r", str(version))
                continue
            if version.is_prerelease:
                log.info("Ignoring pre-release %r", str(version))
                continue
            versions.append(VersionTuple(version, version_string))
        except InvalidVersion:
            log.warning("Ignoring release %r", version_string)
    versions.sort(reverse=True)
    return versions


@click.command()
@click.option("--force", is_flag=True)
@click.option(
    "--level", default="INFO", type=click.Choice(["WARNING", "INFO", "DEBUG"], case_sensitive=False)
)
def update_nodejs(force: bool, level: str) -> None:
    logging.basicConfig(
        level="INFO",
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True, markup=True)],
    )
    log.setLevel(level)

    toml_file_path = RESOURCES_DIR / "nodejs.toml"

    original_toml = toml_file_path.read_text()
    with toml_file_path.open("rb") as f:
        nodejs_data = tomllib.load(f)

    nodejs_data.pop("url")

    major_versions = [VersionTuple(Version(key), key) for key in nodejs_data]
    major_versions.sort(reverse=True)

    versions = parse_nodejs_index()

    # update existing versions, 1 per LTS
    for major_version in major_versions:
        current = Version(nodejs_data[major_version.version_string])
        specifier = packaging.specifiers.SpecifierSet(
            specifiers=f"=={major_version.version.major}.*"
        )
        for version in versions:
            if specifier.contains(version.version) and version.version > current:
                nodejs_data[major_version.version_string] = version.version_string
                break

    # check for a new major LTS to insert
    if versions and versions[0].version.major > major_versions[0].version.major:
        major_versions.insert(
            0,
            VersionTuple(Version(str(versions[0].version.major)), f"v{versions[0].version.major}"),
        )
        nodejs_data[major_versions[0].version_string] = versions[0].version_string

    versions_toml = "\n".join(
        f'{major_version.version_string} = "{nodejs_data[major_version.version_string]}"'
        for major_version in major_versions
    )
    result_toml = f'url = "{NODEJS_DIST}"\n{versions_toml}\n'

    rich.print()  # spacer

    if original_toml == result_toml:
        rich.print("[green]Check complete, nodejs version unchanged.")
        return

    rich.print("nodejs version updated.")
    rich.print("Changes:")
    rich.print()

    toml_relpath = toml_file_path.relative_to(DIR).as_posix()
    diff_lines = difflib.unified_diff(
        original_toml.splitlines(keepends=True),
        result_toml.splitlines(keepends=True),
        fromfile=toml_relpath,
        tofile=toml_relpath,
    )
    rich.print(Syntax("".join(diff_lines), "diff", theme="ansi_light"))
    rich.print()

    if force:
        toml_file_path.write_text(result_toml)
        rich.print("[green]TOML file updated.")
    else:
        rich.print("[yellow]File left unchanged. Use --force flag to update.")


if __name__ == "__main__":
    update_nodejs()



#!/usr/bin/env python3


import argparse
import os
import subprocess
import sys
import tempfile
from pathlib import Path

if __name__ == "__main__":
    # move cwd to the project root
    os.chdir(Path(__file__).resolve().parents[1])

    parser = argparse.ArgumentParser(description="Runs a sample build", allow_abbrev=False)
    parser.add_argument("project_python_path", nargs="?", default="test.test_0_basic.basic_project")

    options = parser.parse_args()

    project_dir = tempfile.mkdtemp()
    subprocess.run(
        [sys.executable, "-m", "test.test_projects", options.project_python_path, project_dir],
        check=True,
    )

    sys.exit(
        subprocess.run(
            [sys.executable, "-m", "cibuildwheel"], cwd=project_dir, check=False
        ).returncode
    )
